import abc
import numpy as _np


class AbstractMachine(abc.ABC):
    """Abstract class for NetKet machines"""

    def __init__(self, hilbert, dtype=complex):
        super().__init__()
        self.hilbert = hilbert

        if dtype is not float and dtype is not complex:
            raise TypeError("dtype must be either float or complex")

        self._dtype = dtype

    @abc.abstractmethod
    def log_val(self, x, out=None):
        r"""Computes the logarithm of the wave function for a batch of visible
        configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. The length of `out` should be `x.shape[0]`.

        Returns:
            A complex number when `x` is a vector and vector when `x` is a
            matrix.
        """
        pass

    def init_random_parameters(self, seed=None, sigma=0.01):
        rgen = _np.random.RandomState(seed)
        self.parameters = rgen.normal(
            scale=sigma, size=self.n_par
        ) + 1.0j * rgen.normal(scale=sigma, size=self.n_par)

    def vector_jacobian_prod(
        self, x, vec, out=None, conjugate=True, return_jacobian=False
    ):
        r"""Computes the scalar product between gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and a vector `vec`. The result is stored into `out`.

        Args:
            x: a matrix or 3d tensor of `float64` of shape `(*, self.n_visible)` or `(*, *, self.n_visible)`.
            vec: a `complex128` vector or matrix used to compute the inner product with the jacobian.
            out: The result of the inner product, it is a vector of `complex128` and length `self.n_par`.
            conjugate (bool): If true, this computes the conjugate of the vector jacobian product.
            return_jacobian (bool): If true, the Jacobian is returned.


        Returns:
            `out` or (out,jacobian) if return_jacobian is True
        """
        vec = vec.reshape(-1)

        if x.ndim == 3:

            jacobian = _np.stack([self.der_log(xb) for xb in x])

            if conjugate:
                out = _np.tensordot(vec, jacobian.conjugate(), axes=1)
            else:
                out = _np.tensordot(vec.conjugate(), jacobian, axes=1)

        elif x.ndim == 2:

            jacobian = self.der_log(x)

            if conjugate:
                out = _np.dot(jacobian.transpose().conjugate(), vec, out)
            else:
                out = _np.dot(jacobian.transpose(), vec.conjugate(), out)

        out = out.reshape(-1)

        return (out, jacobian) if return_jacobian else out

    def jacobian_vector_prod(
        self, x, vec, out=None, conjugate=True, return_jacobian=False
    ):
        raise NotImplementedError

    def der_log(self, x, out=None):
        r"""Computes the gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination tensor of `complex128`.
                `out` should be a matrix of shape `(v.shape[0], self.n_par)`.

        Returns:
            `out`
        """
        raise NotImplementedError

    def to_array(self, normalize=True, batch_size=512):
        r"""
        Returns a numpy array representation of the machine.
        Note that, in general, the size of the array is exponential
        in the number of quantum numbers, and this operation should thus
        only be performed for low-dimensional Hilbert spaces.

        This method requires an indexable Hilbert space.

        Args:
            normalize (bool): If True, the returned array is normalized in L2 norm.
            batch_size (int): The method log_val of the machine is called on
                              batch_size states at once.

        Returns:
            numpy.array: The array machine(x) for all states x in the Hilbert space.

        """
        if self.hilbert.is_indexable:
            psi = self.log_val(self.hilbert.all_states())
            logmax = psi.real.max()
            psi = _np.exp(psi - logmax)

            if normalize:
                norm = _np.linalg.norm(psi)
                psi /= norm

            return psi
        else:
            raise RuntimeError("The hilbert space is not indexable")

    def log_norm(self, order=2):
        r"""
        Returns the log of the norm of the machine.
        Note that, in general, the size of the array is exponential
        in the number of quantum numbers, and this operation should thus
        only be performed for low-dimensional Hilbert spaces.

        This method requires an indexable Hilbert space.

        Args:
            order (int): By default order=2 is the L2 norm.

        Returns:
            float: log(|machine(x)|^order)

        """
        if self.hilbert.is_indexable:
            psi = self.log_val(self.hilbert.all_states())
            maxl = psi.real.max()
            log_n = _np.log(_np.exp(order * (psi.real - maxl)).sum())

            return log_n + maxl * order
        else:
            raise RuntimeError("The hilbert space is not indexable")

    @property
    def dtype(self):
        return self._dtype

    @property
    def has_complex_parameters(self):
        return self._dtype is complex

    @property
    def n_par(self):
        r"""The number of variational parameters in the machine."""
        return self.parameters.size

    @property
    @abc.abstractmethod
    def state_dict(self):
        r"""A dictionary containing the parameters of this machine"""
        raise NotImplementedError

    @property
    def parameters(self):
        return _np.concatenate(tuple(p.reshape(-1) for p in self.state_dict.values()))

    @parameters.setter
    def parameters(self, p):
        if p.shape != (self.n_par,):
            raise ValueError(
                "p has wrong shape: {}; expected ({},)".format(p.shape, self.n_par)
            )

        i = 0
        for x in map(lambda x: x.reshape(-1), self.state_dict.values()):
            _np.copyto(x, p[i : i + x.size])
            i += x.size

    def numpy_flatten(self, data):
        r"""Returns a flattened numpy array representing the given data.
            This is typically used to serialize parameters and gradients.
            The default implementation attempts to return a simple reshaped view.

        Args:
            data: a contigous numpy-compatible array.

        Returns:
            numpy.ndarray: a one-dimensional array containing a view of the data
        """
        return _np.asarray(data).reshape(-1)

    def numpy_unflatten(self, data, shape_like):
        r"""Attempts a deserialization of the given numpy data.
            This is typically used to deserialize parameters and gradients.

        Args:
            data: a 1d numpy array.
            shape_like: this as in instance having the same type and shape of the desired conversion.

        Returns:
            A numpy array containing a view of data and compatible with the given shape.
        """
        return _np.asarray(data).reshape(shape_like.shape)

    def save(self, file):
        assert type(file) is str
        with open(file, "wb") as file_ob:
            _np.save(file_ob, self.numpy_flatten(self.parameters), allow_pickle=False)

    def load(self, file):
        self.parameters = self.numpy_unflatten(
            _np.load(file, allow_pickle=False), self.parameters
        )

    @property
    def input_size(self):
        return self.hilbert.size

    def __repr__(self):
        return "{}(input_size={}, n_par={}, complex_parameters={})".format(
            type(self).__name__,
            self.input_size,
            self.n_par,
            self.has_complex_parameters,
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk

# 1D Lattice
L = 20  # 10

g = nk.graph.Hypercube(length=L, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)

# Ising spin Hamiltonian
ha = nk.operator.Ising(hilbert=hi, graph=g, h=1)

# Autoregressive neural network
ma = nk.models.ARNNConv1D(hilbert=hi, layers=2, features=10, kernel_size=10)

# Autoregressive direct sampling
sa = nk.sampler.ARDirectSampler(hi)

# Optimizer
op = nk.optimizer.Sgd(learning_rate=0.1)

# SR
sr = nk.optimizer.SR(diag_shift=0.01)

# Variational state
# With direct sampling, we don't need many samples in each step to form a
# Markov chain, and we don't need to discard samples
vs = nk.vqs.MCState(sa, ma, n_samples=64)

# n_parameters also takes masked parameters into account
# The number of non-masked parameters is about a half
print("n_parameters:", vs.n_parameters)

# Variational monte carlo driver with a variational state
gs = nk.VMC(ha, op, variational_state=vs, preconditioner=sr)

# Run the optimization for 1000 iterations
gs.run(n_iter=1000, out="test")

# Use more samples to estimate the Hamiltonian and the error bar
vs.n_samples = 1000
print(vs.expect(ha))



import numpy as _np
from ._sum_inplace import sum_inplace as _sum_inplace

from netket.utils.mpi import (
    mpi_available as _mpi_available,
    n_nodes as _n_nodes,
    MPI_py_comm as MPI_comm,
)

if _mpi_available:
    from netket.utils.mpi import MPI


def subtract_mean(x, axis=None):
    """
    Subtracts the mean of the input array over all but the last dimension
    and over all MPI processes from each entry.

    Args:
        x: Input array
        axis: Axis or axes along which the means are computed. The default (None) is to
              compute the mean of the flattened array.

    Returns:
        The resulting array.

    """
    x_mean = mean(x, axis=axis)
    x -= x_mean

    return x


def mean(a, axis=None, keepdims: bool = False):
    """
    Compute the arithmetic mean along the specified axis and over MPI processes.

    Returns the average of the array elements. The average is taken over the flattened array by default,
    otherwise over the specified axis. float64 intermediate and return values are used for integer inputs.

    Args:
        a: The input array
        axis: Axis or axes along which the means are computed. The default (None) is to
              compute the mean of the flattened array.
        keepdims: If True the output array will have the same number of dimensions as the input,
              with the reduced axes having length 1. (default=False)

    Returns:
        The array with reduced dimensions defined by axis.

    """
    out = a.mean(axis=axis, keepdims=keepdims)

    out = _sum_inplace(out)
    out /= _n_nodes

    return out


def sum(a, axis=None, out=None, keepdims: bool = False):
    """
    Compute the sum along the specified axis and over MPI processes.

    Args:
        a: The input array
        axis: Axis or axes along which the mean is computed. The default (None) is to
              compute the mean of the flattened array.
        out: An optional pre-allocated array to fill with the result.
        keepdims: If True the output array will have the same number of dimensions as the input,
              with the reduced axes having length 1. (default=False)

    Returns:
        The array with reduced dimensions defined by axis. If out is not none, returns out.

    """
    # asarray is necessary for the axis=None case to work, as the MPI call requires a NumPy array
    out = _np.asarray(_np.sum(a, axis=axis, out=out, keepdims=keepdims))

    if _n_nodes > 1:
        MPI_comm.Allreduce(MPI.IN_PLACE, out.reshape(-1), op=MPI.SUM)

    return out


def var(a, axis=None, out=None, ddof: int = 0):
    """
    Compute the variance mean along the specified axis and over MPI processes.

    Args:
        a: The input array
        axis: Axis or axes along which the variance is computed. The default (None) is to
              compute the variance of the whole flattened array.
        out: An optional pre-allocated array to fill with the result.
        ddof: “Delta Degrees of Freedom”: the divisor used in the calculation is N - ddof,
              where N represents the number of elements. By default ddof is zero.

    Returns:
        The array with reduced dimensions defined by axis. If out is not none, returns out.

    """
    m = mean(a, axis=axis)

    if axis is None:
        ssq = _np.abs(a - m) ** 2.0
    else:
        ssq = _np.abs(a - _np.expand_dims(m, axis)) ** 2.0

    out = sum(ssq, axis=axis, out=out)

    n_all = total_size(a, axis=axis)
    out /= n_all - ddof

    return out


def total_size(a, axis=None):
    """
    Compute the total number of elements stored in the input array among all MPI processes.

    This function essentially returns MPI_sum_among_processes(a.size).

    Args:
        a: The input array.
        axis: If specified, only considers the total size of that axis.

    Returns:
        a.size or a.shape[axis], reduced among all MPI processes.
    """
    if axis is None:
        l_size = a.size
    else:
        l_size = a.shape[axis]

    if _n_nodes > 1:
        l_size = MPI_comm.allreduce(l_size, op=MPI.SUM)

    return l_size



from functools import singledispatch
from . import numpy


@singledispatch
def Sgd(machine, learning_rate, l2reg=0, decay_factor=1.0):
    r"""Stochastic Gradient Descent Optimizer.
    The `Stochastic Gradient Descent <https://en.wikipedia.org/wiki/Stochastic_gradient_descent>`_
    is one of the most popular optimizers in machine learning applications.
    Given a stochastic estimate of the gradient of the cost function (:math:`G(\mathbf{p})`),
    it performs the update:

    .. math:: p^\prime_k = p_k -\eta G_k(\mathbf{p}),

    where :math:`\eta` is the so-called learning rate.
    NetKet also implements two extensions to the simple SGD,
    the first one is :math:`L_2` regularization,
    and the second one is the possibility to set a decay
    factor :math:`\gamma \leq 1` for the learning rate, such that
    at iteration :math:`n` the learning rate is :math:`\eta \gamma^n`.

    Args:
       machine (netket.AbstractMachine): The machine whose parameters are being optimized.
       learning_rate (float): The learning rate :math:`\eta`.
       l2_reg (float): The amount of :math:`L_2` regularization.
       decay_factor (float): The decay factor :math:`\gamma`.

    Examples:
       Simple SGD optimizer.

       >>> from netket.optimizer import Sgd
       >>> op = Sgd(learning_rate=0.05)
    """
    return numpy.Sgd(learning_rate, l2reg, decay_factor)



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import numpy as jnp
import flax
from flax import linen as nn
from functools import partial

from .jax import Jax, add_package_wrapper

from jax.tree_util import tree_flatten, tree_unflatten, tree_map

from ._jax_utils import forward_apply, grad, vjp, tree_leaf_iscomplex


def flax_init(module, rng, shape):
    shape = (1,) + shape[1:]
    dummy_input = jax.numpy.zeros(shape, dtype=jax.numpy.float64)
    params = module.init(rng, dummy_input)
    return (-1, 1), params


# Declare Flax as a jax wrapper package.
add_package_wrapper(
    lambda m: isinstance(m, flax.linen.Module),
    lambda m: (partial(flax_init, m), m.apply),
)


class Flax(Jax):
    def __init__(self, hilbert, module):
        """
        Wraps a stax network (which is a tuple of `init_fn` and `predict_fn`)
        so that it can be used as a NetKet machine.

        Args:
            hilbert: Hilbert space on which the state is defined. Should be a
                subclass of `netket.hilbert.Hilbert`.
            module: A pair `(init_fn, predict_fn)`. See the documentation of
                jax.experimental.stax` for more info.
            dtype: either complex or float, is the type used for the weights.
                In both cases the network must have a single output.
        """
        super().__init__(hilbert=hilbert, module=module)
        self._dtype = complex if tree_leaf_iscomplex(self._params) else float

    def init_random_parameters(self, seed=None, sigma=None):
        self.jax_init_parameters(seed)

    def _cast(self, p):
        return p



from .metropolis_hastings import *
from ._kernels import _CustomKernel


def CustomSampler(
    machine, move_operators, move_weights=None, n_chains=16, sweep_size=None, **kwargs
):
    r"""
     Custom Sampler, where transition operators are specified by the user.
     For the moment, this functionality is limited to transition operators which
     are sums of :math:`k`-local operators:

     .. math::
        \mathcal{M}= \sum_i M_i


     where the move operators :math:`M_i` act on an (arbitrary) subset of sites.

     The operators :math:`M_i` are specified giving their matrix elements, and a list
     of sites on which they act. Each operator :math:`M_i` must be real,
     symmetric, positive definite and stochastic (i.e. sum of each column and line is 1).

     The transition probability associated to a custom sampler can be decomposed into two steps:

     1. One of the move operators :math:`M_i` is chosen with a weight given by the
       user (or uniform probability by default). If the weights are provided,
       they do not need to sum to unity.

     2. Starting from state
       :math:`|n \rangle`, the probability to transition to state
       :math:`|m\rangle` is given by
       :math:`\langle n|  M_i | m \rangle`.

     Args:
        machine: A machine :math:`\Psi(s)` used for the sampling.
               The probability distribution being sampled
               from is :math:`F(\Psi(s))`, where the function
               :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
        move_operators: The stochastic `LocalOperator`
             :math:`\mathcal{M}= \sum_i M_i` used for transitions.
        move_weights: For each :math:`i`, the probability to pick one of
             the move operators (must sum to one).
        n_chains: The number of Markov Chains to be run in parallel on a single process.
        sweep_size: The number of exchanges that compose a single sweep.
                    If None, sweep_size is equal to the number of degrees of freedom (n_visible).

    Examples:
        Sampling from a RBM machine in a 1D lattice of spin 1/2

        >>> import netket as nk
        >>>
        >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
        >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
        >>>
        >>> # RBM Spin Machine
        >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
        >>>
        >>> # Construct a Custom Sampler
        >>> # Using random local spin flips (Pauli X operator)
        >>> X = [[0, 1],[1, 0]]
        >>> move_op = nk.operator.LocalOperator(hilbert=hi,operators=[X] * g.n_nodes, acting_on=[[i] for i in range(g.n_nodes)])
        >>> sa = nk.sampler.CustomSampler(machine=ma, move_operators=move_op)
    """
    return MetropolisHastings(
        machine,
        _CustomKernel(machine, move_operators, move_weights),
        n_chains,
        sweep_size,
        **kwargs,
    )


def CustomSamplerPt(
    machine, move_operators, move_weights=None, n_replicas=16, sweep_size=None, **kwargs
):
    r"""
    This sampler performs parallel-tempering
    moves in addition to the local moves implemented in `CustomSampler`.
    The number of replicas can be chosen by the user.

    Args:
      machine: A machine :math:`\Psi(s)` used for the sampling.
               The probability distribution being sampled
               from is :math:`F(\Psi(s))`, where the function
               :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
      move_operators: The stochastic `LocalOperator`
               :math:`\mathcal{M}= \sum_i M_i` used for transitions.
      move_weights: For each :math:`i`, the probability to pick one of
               the move operators (must sum to one).
      n_replicas: The number of replicas used for parallel tempering.
      sweep_size: The number of exchanges that compose a single sweep.
                  If None, sweep_size is equal to the number of degrees of freedom (n_visible).
    """
    return MetropolisHastingsPt(
        machine,
        _CustomKernel(machine, move_operators, move_weights),
        n_replicas,
        sweep_size,
        **kwargs,
    )



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import random

from .abstract_density_matrix import AbstractDensityMatrix
from ..jax import Jax as JaxPure, logcosh
from functools import partial


class Jax(JaxPure, AbstractDensityMatrix):
    def __init__(self, hilbert, module, dtype=complex):
        """
        Wraps a stax network (which is a tuple of `init_fn` and `predict_fn`)
        so that it can be used as a NetKet density matrix.

        Args:
            hilbert: Hilbert space on which the state is defined. Should be a
                subclass of `netket.hilbert.Hilbert`.
            module: A pair `(init_fn, predict_fn)`. See the documentation of
                `jax.experimental.stax` for more info.
            dtype: either complex or float, is the type used for the weights.
                In both cases the module must have a single output.
        """
        AbstractDensityMatrix.__init__(self, hilbert, dtype)
        JaxPure.__init__(self, hilbert, module, dtype)

        assert self.input_size == self.hilbert.size * 2

    @staticmethod
    @jax.jit
    def _dminput(xr, xc):
        if xc is None:
            x = xr
        else:
            x = jnp.hstack((xr, xc))
        return x

    def log_val(self, xr, xc=None, out=None):
        x = self._dminput(xr, xc)

        return JaxPure.log_val(self, x, out=out)

    def der_log(self, xr, xc=None, out=None):
        x = self._dminput(xr, xc)

        return JaxPure.der_log(self, x, out=out)

    def diagonal(self):
        from .diagonal import Diagonal

        diag = Diagonal(self)

        def diag_jax_forward(params, x):
            return self.jax_forward(params, self._dminput(x, x))

        diag.jax_forward = diag_jax_forward

        return diag


from jax.experimental import stax
from jax.experimental.stax import Dense
from jax.nn.initializers import glorot_normal, normal


def DensePurificationComplex(
    out_pure, out_mix, use_hidden_bias=True, W_init=glorot_normal(), b_init=normal()
):
    """Layer constructor function for a complex purification layer."""

    def init_fun(rng, input_shape):
        assert input_shape[-1] % 2 == 0
        output_shape = input_shape[:-1] + (2 * out_pure + out_mix,)

        k = jax.random.split(rng, 7)

        input_size = input_shape[-1] // 2

        # Weights for the pure part
        Wr, Wi = (
            W_init(k[0], (input_size, out_pure)),
            W_init(k[1], (input_size, out_pure)),
        )

        # Weights for the mixing part
        Vr, Vi = (
            W_init(k[2], (input_size, out_mix)),
            W_init(k[3], (input_size, out_mix)),
        )

        if use_hidden_bias:
            br, bi = (b_init(k[4], (out_pure,)), b_init(k[5], (out_pure,)))
            cr = b_init(k[6], (out_mix,))

            return output_shape, (Wr, Wi, Vr, Vi, br, bi, cr)
        else:
            return output_shape, (Wr, Wi, Vr, Vi)

    @jax.jit
    def apply_fun(params, inputs, **kwargs):
        if use_hidden_bias:
            Wr, Wi, Vr, Vi, br, bi, cr = params
        else:
            Wr, Wi, Vr, Vi = params

        xr, xc = jax.numpy.split(inputs, 2, axis=-1)

        thetar = jax.numpy.dot(
            xr[
                :,
            ],
            (Wr + 1.0j * Wi),
        )
        thetac = jax.numpy.dot(
            xc[
                :,
            ],
            (Wr - 1.0j * Wi),
        )

        thetam = jax.numpy.dot(
            xr[
                :,
            ],
            (Vr + 1.0j * Vi),
        )
        thetam += jax.numpy.dot(
            xc[
                :,
            ],
            (Vr - 1.0j * Vi),
        )

        if use_hidden_bias:
            thetar += br + 1.0j * bi
            thetac += br - 1.0j * bi
            thetam += 2 * cr

        return jax.numpy.hstack((thetar, thetam, thetac))

    return init_fun, apply_fun


from ..jax import LogCoshLayer, SumLayer


def ndmSpin(hilbert, alpha, beta, use_hidden_bias=True):
    return stax.serial(
        DensePurificationComplex(
            int(alpha * hilbert.size), int(beta * hilbert.size), use_hidden_bias
        ),
        LogCoshLayer,
        SumLayer,
    )


def NdmSpin(hilbert, alpha, beta, use_hidden_bias=True):
    r"""
    A fully connected Neural Density Matrix (DBM). This type density matrix is
    obtained purifying a RBM with spin 1/2 hidden units.

    The number of purification hidden units can be chosen arbitrarily.

    The weights are taken to be complex-valued. A complete definition of this
    machine can be found in Eq. 2 of Hartmann, M. J. & Carleo, G.,
    Phys. Rev. Lett. 122, 250502 (2019).

    Args:
        hilbert: Hilbert space of the system.
        alpha: `alpha * hilbert.size` is the number of hidden spins used for
                the pure state part of the density-matrix.
        beta: `beta * hilbert.size` is the number of hidden spins used for the purification.
            beta=0 for example corresponds to a pure state.
        use_hidden_bias: If ``True`` bias on the hidden units is taken.
                         Default ``True``.
    """
    return Jax(
        hilbert,
        ndmSpin(hilbert, alpha, beta, use_hidden_bias),
        dtype=float,
    )


def DenseMixingReal(
    out_mix, use_hidden_bias=True, W_init=glorot_normal(), b_init=normal()
):
    """Layer constructor function for a complex purification layer."""

    def init_fun(rng, input_shape):
        assert input_shape[-1] % 2 == 0
        output_shape = input_shape[:-1] + (out_mix,)

        k = jax.random.split(rng, 3)

        input_size = input_shape[-1] // 2

        # Weights for the mixing part
        Ur, Ui = (
            W_init(k[0], (input_size, out_mix)),
            W_init(k[1], (input_size, out_mix)),
        )

        if use_hidden_bias:
            dr = b_init(k[2], (out_mix,))

            return output_shape, (Ur, Ui, dr)
        else:
            return output_shape, (Ur, Ui)

    @jax.jit
    def apply_fun(params, inputs, **kwargs):
        if use_hidden_bias:
            Ur, Ui, dr = params
        else:
            Ur, Ui = params

        xr, xc = jax.numpy.split(inputs, 2, axis=-1)

        theta = jax.numpy.dot(
            xr[
                :,
            ],
            (0.5 * Ur + 0.5j * Ui),
        )
        theta += jax.numpy.dot(
            xc[
                :,
            ],
            (0.5 * Ur - 0.5j * Ui),
        )

        if use_hidden_bias:
            theta += dr

        return theta

    return init_fun, apply_fun


def DensePureRowCol(
    out_pure, use_hidden_bias=True, W_init=glorot_normal(), b_init=normal()
):
    def init_fun(rng, input_shape):
        assert input_shape[-1] % 2 == 0
        input_size = input_shape[-1] // 2

        single_output_shape = input_shape[:-1] + (out_pure,)
        output_shape = (single_output_shape, single_output_shape)

        k = jax.random.split(rng, 3)

        W = W_init(k[0], (input_size, out_pure))

        if use_hidden_bias:
            b = b_init(k[2], (out_pure,))

            return output_shape, (W, b)
        else:
            return output_shape, (W,)

    def apply_fun(params, inputs, **kwargs):
        if use_hidden_bias:
            W, b = params
        else:
            W = params

        xr, xc = jax.numpy.split(inputs, 2, axis=-1)

        thetar = jax.numpy.dot(
            xr[
                :,
            ],
            W,
        )
        thetac = jax.numpy.dot(
            xc[
                :,
            ],
            W,
        )

        if use_hidden_bias:
            thetar += b
            thetac += b

        return (thetar, thetac)

    return init_fun, apply_fun


def FanInSum2():
    """Layer construction function for a fan-in sum layer."""

    def init_fun(rng, input_shape):
        output_shape = input_shape[0]
        return output_shape, tuple()

    def apply_fun(params, inputs, **kwargs):
        output = 0.5 * (inputs[0] + inputs[1])
        return output

    return init_fun, apply_fun


FanInSum2 = FanInSum2()


def FanInSub2():
    """Layer construction function for a fan-in sum layer."""

    def init_fun(rng, input_shape):
        output_shape = input_shape[0]
        return output_shape, tuple()

    def apply_fun(params, inputs, **kwargs):
        output = 0.5j * (inputs[0] - inputs[1])
        return output

    return init_fun, apply_fun


FanInSub2 = FanInSub2()


def BiasRealModPhase(b_init=normal()):
    def init_fun(rng, input_shape):
        assert input_shape[-1] % 2 == 0
        input_size = input_shape[-1] // 2

        output_shape = input_shape[:-1]

        k = jax.random.split(rng, 2)

        br = b_init(k[0], (input_size,))
        bj = b_init(k[1], (input_size,))

        return output_shape, (br, bj)

    def apply_fun(params, inputs, **kwargs):
        br, bj = params

        xr, xc = jax.numpy.split(inputs, 2, axis=-1)

        biasr = jax.numpy.dot(
            (xr + xc)[
                :,
            ],
            br,
        )
        biasj = jax.numpy.dot(
            (xr - xc)[
                :,
            ],
            bj,
        )

        return 0.5 * biasr + 0.5j * biasj

    return init_fun, apply_fun


def ndmSpinPhase(hilbert, alpha, beta, use_hidden_bias=True, use_visible_bias=True):
    mod_pure = stax.serial(
        DensePureRowCol(alpha * hilbert.size, use_hidden_bias),
        stax.parallel(LogCoshLayer, LogCoshLayer),
        stax.parallel(SumLayer, SumLayer),
        FanInSum2,
    )

    phs_pure = stax.serial(
        DensePureRowCol(alpha * hilbert.size, use_hidden_bias),
        stax.parallel(LogCoshLayer, LogCoshLayer),
        stax.parallel(SumLayer, SumLayer),
        FanInSub2,
    )

    mixing = stax.serial(
        DenseMixingReal(int(beta * hilbert.size), use_hidden_bias),
        LogCoshLayer,
        SumLayer,
    )

    if use_visible_bias:
        biases = BiasRealModPhase()
        net = stax.serial(
            stax.FanOut(4),
            stax.parallel(mod_pure, phs_pure, mixing, biases),
            stax.FanInSum,
        )
    else:
        net = stax.serial(
            stax.FanOut(3),
            stax.parallel(mod_pure, phs_pure, mixing),
            stax.FanInSum,
        )

    return net


def NdmSpinPhase(hilbert, alpha, beta, use_hidden_bias=True, use_visible_bias=True):
    r"""
    A fully connected Neural Density Matrix (DBM). This type density matrix is
    obtained purifying a RBM with spin 1/2 hidden units.

    The number of purification hidden units can be chosen arbitrarily.

    The weights are taken to be complex-valued. A complete definition of this
    machine can be found in Eq. 2 of Hartmann, M. J. & Carleo, G.,
    Phys. Rev. Lett. 122, 250502 (2019).

    Args:
        hilbert: Hilbert space of the system.
        alpha: `alpha * hilbert.size` is the number of hidden spins used for
                the pure state part of the density-matrix.
        beta: `beta * hilbert.size` is the number of hidden spins used for the purification.
            beta=0 for example corresponds to a pure state.
        use_hidden_bias: If ``True`` bias on the hidden units is taken.
                         Default ``True``.
    """
    return Jax(
        hilbert,
        ndmSpinPhase(hilbert, alpha, beta, use_hidden_bias, use_visible_bias),
        dtype=float,
    )


def JaxRbmSpin(hilbert, alpha, dtype=complex):
    return Jax(
        hilbert,
        stax.serial(stax.Dense(alpha * hilbert.size * 2), LogCoshLayer, SumLayer),
        dtype=dtype,
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa

import json
import dataclasses
import orjson

from os import path as _path
import numpy as np
import jax

from flax import serialization

from jax.tree_util import tree_map


def _exists_json(prefix):
    return _path.exists(prefix + ".log") or _path.exists(prefix + ".mpack")


def default(obj):
    if hasattr(obj, "to_json"):
        return obj.to_json()
    elif isinstance(obj, np.ndarray):
        if obj.ndim == 0:
            return obj.item()
        elif obj.ndim == 1:
            return obj.tolist()
        else:
            raise TypeError
    elif hasattr(obj, "_device"):
        return np.array(obj)
    elif isinstance(obj, complex):
        return obj.real

    raise TypeError


class JsonLog:
    """
    Creates a Json Logger sink object, that can be passed with keyword argument `logger` to Monte
    Carlo drivers in order to serialize the outpit data of the simulation.

    If the model state is serialized, then it is serialized using the msgpack protocol of flax.
    For more information on how to de-serialize the output, see
    `here <https://flax.readthedocs.io/en/latest/flax.serialization.html>`_

    Args:
        output_prefix: the name of the output files before the extension
        save_params_every: every how many iterations should machine parameters be flushed to file
        write_every: every how many iterations should data be flushed to file
        mode: Specify the behaviour in case the file already exists at this output_prefix. Options
        are
        - `[w]rite`: (default) overwrites file if it already exists;
        - `[a]ppend`: appends to the file if it exists, overwise creates a new file;
        - `[x]` or `fail`: fails if file already exists;
    """

    def __init__(
        self, output_prefix, mode="write", save_params_every=50, write_every=50
    ):
        # Shorthands for mode
        if mode == "w":
            mode = "write"
        elif mode == "a":
            mode = "append"
        elif mode == "x":
            mode = "fail"

        if not ((mode == "write") or (mode == "append") or (mode == "fail")):
            raise ValueError(
                "Mode not recognized: should be one of `[w]rite`, `[a]ppend` or `[x]`(fail)."
            )

        file_exists = _exists_json(output_prefix)

        starting_json_content = {"Output": []}

        if file_exists and mode == "append":
            # if there is only the .mpacck file but not the json one, raise an error
            if not _path.exists(output_prefix + ".log"):
                raise ValueError(
                    "History file does not exists, but wavefunction file does. Please change `output_prefix or set mode=`write`."
                )

            starting_json_content = json.load(open(output_prefix + ".log"))

        elif file_exists and mode == "fail":
            raise ValueError(
                "Output file already exists. Either delete it manually or change `output_prefix`."
            )

        self._json_out = starting_json_content
        self._prefix = output_prefix
        self._write_every = write_every
        self._save_params_every = save_params_every
        self._old_step = 0
        self._steps_notflushed_write = 0
        self._steps_notflushed_pars = 0

    def __call__(self, step, item, machine):
        item["Iteration"] = step

        self._json_out["Output"].append(item)

        if (
            self._steps_notflushed_write % self._write_every == 0
            or step == self._old_step - 1
        ):
            self._flush_log()
        if (
            self._steps_notflushed_pars % self._save_params_every == 0
            or step == self._old_step - 1
        ):
            self._flush_params(machine)

        self._old_step = step
        self._steps_notflushed_write += 1
        self._steps_notflushed_pars += 1

    def _flush_log(self):
        with open(self._prefix + ".log", "wb") as outfile:

            outfile.write(orjson.dumps(self._json_out, default=default))
            self._steps_notflushed_write = 0

    def _flush_params(self, variational_state):
        binary_data = serialization.to_bytes(variational_state.variables)
        with open(self._prefix + ".mpack", "wb") as outfile:
            outfile.write(binary_data)

        self._steps_notflushed_pars = 0

    def flush(self, variational_state):
        """
        Writes to file the content of this logger.

        :param machine: optionally also writes the parameters of the machine.
        """
        self._flush_log()

        if variational_state is not None:
            self._flush_params(variational_state)



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from functools import partial

import numpy as _np
from jax import numpy as jnp
from jax.tree_util import (
    tree_flatten,
    tree_unflatten,
    tree_map,
    tree_multimap,
    tree_leaves,
)


def forward_scalar(pars, forward_fn, x):
    """
    Performs the forward pass of a neural network on a single state (vector) input.
    """
    return forward_fn(pars, jnp.expand_dims(x, 0)).reshape(())


forward_apply = jax.jit(
    lambda pars, forward_fn, x: forward_fn(pars, x), static_argnums=1
)

##
@jax.jit
def tree_size(tree):
    """
    Returns the sum of the size of all leaves in the tree.
    It's equivalent to the number of scalars in the pytree.
    """
    return sum(tree_leaves(tree_map(lambda x: x.size, tree)))


def tree_leaf_iscomplex(pars):
    """
    Returns true if at least one leaf in the tree has complex dtype.
    """

    def _has_complex_dtype(x):
        #  Returns true if x is complex
        return jnp.issubdtype(x.dtype, jnp.complexfloating)

    return any(jax.tree_leaves(jax.tree_map(_has_complex_dtype, pars)))


def outdtype(forward_fn, pars, v):
    """
    Returns the dtype of forward_fn(pars, v)
    """
    if v.ndim > 1:
        v = v.reshape(-1, v.shape[-1])[0, :]
    return forward_scalar(pars, forward_fn, v).dtype


def outdtype_iscomplex(forward_fn, pars, v):
    """
    Returns true if forward_fn(pars, v) is complex
    """
    return jnp.issubdtype(outdtype(forward_fn, pars, v), jnp.complexfloating)


# _grad_CC, _RR and _RC are the batched gradient functions for machines going
# from R -> C, R->R and R->C. Ditto for vjp
# Thee reason why R->C is more complicated is that it splits the calculation
# into the real and complex part in order to be more efficient.

_grad_CC = jax.grad(forward_scalar, holomorphic=True)
_grad_RR = jax.grad(forward_scalar)


def _grad_RC(pars, forward_fn, v):
    grad_r = jax.grad(
        lambda pars, forward_fn, v: forward_scalar(pars, forward_fn, v).real
    )(pars, forward_fn, v)
    grad_j = jax.grad(
        lambda pars, forward_fn, v: forward_scalar(pars, forward_fn, v).imag
    )(pars, forward_fn, v)

    r_flat, r_fun = tree_flatten(grad_r)
    j_flat, j_fun = tree_flatten(grad_j)

    grad_flat = [re + 1j * im for re, im in zip(r_flat, j_flat)]
    return tree_unflatten(r_fun, grad_flat)


def _grad(pars, forward_fn, v):
    if tree_leaf_iscomplex(pars):
        if outdtype_iscomplex(forward_fn, pars, v):  # C -> C
            return _grad_CC(pars, forward_fn, v)
        else:
            raise RuntimeError("C->R function detected, but not supported.")
    else:
        if outdtype_iscomplex(forward_fn, pars, v):  # R -> C
            return _grad_RC(pars, forward_fn, v)
        else:
            return _grad_RR(pars, forward_fn, v)


grad = jax.jit(jax.vmap(_grad, in_axes=(None, None, 0)), static_argnums=1)


def _vjp_CC(pars, forward_fn, v, vec, conjugate):
    vals, f_vjp = jax.vjp(forward_fn, pars, v.reshape((-1, v.shape[-1])))

    out = f_vjp(vec.reshape(vals.shape).conjugate())[0]

    if conjugate:
        out = tree_map(jnp.conjugate, out)

    return out


def _vjp_RR(pars, forward_fn, v, vec, conjugate):
    vals, f_jvp = jax.vjp(forward_fn, pars, v.reshape((-1, v.shape[-1])))

    out_r = f_jvp(vec.reshape(vals.shape).real)[0]
    out_i = f_jvp(-vec.reshape(vals.shape).imag)[0]

    if conjugate:
        out = tree_multimap(lambda re, im: re - 1j * im, out_r, out_i)
    else:
        out = tree_multimap(lambda re, im: re + 1j * im, out_r, out_i)

    return out


def _vjp_RC(pars, forward_fn, v, vec, conjugate):
    v = v.reshape((-1, v.shape[-1]))
    vals_r, f_jvp_r = jax.vjp(lambda pars, v: forward_fn(pars, v).real, pars, v)

    vals_j, f_jvp_j = jax.vjp(lambda pars, v: forward_fn(pars, v).imag, pars, v)

    vec_r = vec.reshape(vals_r.shape).real
    vec_j = vec.reshape(vals_r.shape).imag

    # val = vals_r + vals_j
    vr_jr = f_jvp_r(vec_r)[0]
    vj_jr = f_jvp_r(vec_j)[0]
    vr_jj = f_jvp_j(vec_r)[0]
    vj_jj = f_jvp_j(vec_j)[0]

    r = tree_multimap(
        lambda re, im: re - 1j * im,
        vr_jr,
        vj_jr,
    )
    i = tree_multimap(lambda re, im: re - 1j * im, vr_jj, vj_jj)
    out = tree_multimap(lambda re, im: re + 1j * im, r, i)

    if conjugate:
        out = tree_map(jnp.conjugate, out)

    return out


#  This function dispatches to the right
def _vjp(pars, forward_fn, v, vec, conjugate):

    # output dtype
    out_dtype = forward_scalar(pars, forward_fn, v[0, :]).dtype
    out_dtype = _np.promote_types(out_dtype, vec.dtype)
    # convert the sensitivity to right dtype
    vec = jnp.asarray(vec, dtype=out_dtype)

    if tree_leaf_iscomplex(pars):
        if jnp.issubdtype(out_dtype, jnp.complexfloating):  # C -> C
            return _vjp_CC(pars, forward_fn, v, vec, conjugate)
        elif jnp.issubdtype(out_dtype, jnp.floating):  # C -> R
            raise RuntimeError("C->R function detected, but not supported.")
    else:
        if jnp.issubdtype(out_dtype, jnp.complexfloating):  # R -> C
            return _vjp_RC(pars, forward_fn, v, vec, conjugate)
        elif jnp.issubdtype(out_dtype, jnp.floating):  # R -> R
            return _vjp_RR(pars, forward_fn, v, vec, conjugate)


vjp = jax.jit(_vjp, static_argnums=(1, 4))



from ..abstract_sampler import AbstractSampler
from netket.legacy import random as _random

from netket.legacy.stats import sum_inplace as _sum_inplace

import math
import numpy as _np
from numba import jit, int64, float64
from numba.experimental import jitclass


class MetropolisHastings(AbstractSampler):
    def __init__(self, machine, kernel, n_chains=16, sweep_size=None):

        super().__init__(machine, n_chains)

        self.n_chains = n_chains

        self.sweep_size = sweep_size

        self._kernel = kernel

        self.machine_pow = 2.0
        self.reset(True)

    @property
    def n_chains(self):
        return self._n_chains

    @n_chains.setter
    def n_chains(self, n_chains):
        if n_chains < 0:
            raise ValueError("Expected a positive integer for n_chains ")

        self._n_chains = n_chains

        self._state = _np.zeros((n_chains, self._input_size))
        self._state1 = _np.copy(self._state)

        self._log_values = _np.zeros(n_chains, dtype=_np.complex128)
        self._log_values_1 = _np.zeros(n_chains, dtype=_np.complex128)
        self._log_prob_corr = _np.zeros(n_chains)

    @property
    def machine_pow(self):
        return self._machine_pow

    @machine_pow.setter
    def machine_pow(self, m_power):
        if not _np.isscalar(m_power):
            raise ValueError("machine_pow should be a scalar.")
        self._machine_pow = m_power

    @property
    def sweep_size(self):
        return self._sweep_size

    @sweep_size.setter
    def sweep_size(self, sweep_size):
        self._sweep_size = sweep_size if sweep_size != None else self._input_size
        if self._sweep_size < 0:
            raise ValueError("Expected a positive integer for sweep_size ")

    def reset(self, init_random=False):
        if init_random:
            self._kernel.random_state(self._state)
        self._log_values = self.machine.log_val(self._state, out=self._log_values)

        self._accepted_samples = 0
        self._total_samples = 0

    @staticmethod
    @jit(nopython=True)
    def acceptance_kernel(
        state, state1, log_values, log_values_1, log_prob_corr, machine_pow
    ):
        accepted = 0

        for i in range(state.shape[0]):
            prob = _np.exp(
                machine_pow * (log_values_1[i] - log_values[i] + log_prob_corr[i]).real
            )
            assert not math.isnan(prob)

            if prob > _random.uniform(0, 1):
                log_values[i] = log_values_1[i]
                state[i] = state1[i]
                accepted += 1

        return accepted

    def __next__(self):

        _log_val = self.machine.log_val
        _acc_kernel = self.acceptance_kernel
        _state = self._state
        _state1 = self._state1
        _log_values = self._log_values
        _log_values_1 = self._log_values_1
        _log_prob_corr = self._log_prob_corr
        _machine_pow = self._machine_pow
        _t_kernel = self._kernel.transition

        accepted = 0

        for sweep in range(self.sweep_size):

            # Propose a new state using the transition kernel
            _t_kernel(_state, _state1, _log_prob_corr)

            _log_values_1 = _log_val(_state1, out=_log_values_1)

            # Acceptance Kernel
            accepted += _acc_kernel(
                _state,
                _state1,
                _log_values,
                _log_values_1,
                _log_prob_corr,
                _machine_pow,
            )

        self._total_samples += self.sweep_size * self.n_chains
        self._accepted_samples += accepted

        return self._state

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return _sum_inplace(self._accepted_samples) / _sum_inplace(self._total_samples)



from .abstract_machine import AbstractMachine
import torch as _torch
import numpy as _np
import warnings
from netket.utils import backpack_available as _backpack_available

if _backpack_available:
    from netket.utils import backpack


def _get_number_parameters(m):
    r"""Returns total number of variational parameters in a torch.nn.Module."""
    return sum(map(lambda p: p.numel(), _get_differentiable_parameters(m)))


def _get_differentiable_parameters(m):
    r"""Returns total number of variational parameters in a torch.nn.Module."""
    return filter(lambda p: p.requires_grad, m.parameters())


class Torch(AbstractMachine):
    def __init__(self, module, hilbert, dtype=float, use_backpack=False):
        self._module = _torch.jit.load(module) if isinstance(module, str) else module
        self._module.double()
        self._n_par = _get_number_parameters(self._module)
        self._parameters = list(_get_differentiable_parameters(self._module))
        self._use_backpack = use_backpack
        if self._use_backpack:
            if not _backpack_available:
                raise RuntimeError("BackPACK was not successfully imported")
            self._module = backpack.extend(self._module)
        # TODO check that module has input shape compatible with hilbert size
        super().__init__(hilbert, dtype=dtype)

    @property
    def parameters(self):
        return (
            _torch.cat(
                tuple(p.view(-1) for p in _get_differentiable_parameters(self._module))
            )
            .detach()
            .numpy()
            .astype(_np.complex128)
        )

    def assign_beta(self, beta):
        self._module.beta = beta
        return

    def save(self, filename):
        _torch.save(self._module.state_dict(), filename)
        return

    def load(self, filename):
        self._module.load_state_dict(_torch.load(filename))
        return

    @parameters.setter
    def parameters(self, p):
        if not _np.all(p.imag == 0.0):
            warnings.warn(
                "PyTorch machines have real parameters, imaginary part will be discarded"
            )
        torch_pars = _torch.from_numpy(p.real)
        if torch_pars.numel() != self._n_par:
            raise ValueError(
                "p has wrong shape: {}; expected [{}]".format(
                    torch_pars.size(), self._n_par
                )
            )
        i = 0
        for x in map(
            lambda x: x.view(-1), _get_differentiable_parameters(self._module)
        ):
            x.data.copy_(torch_pars[i : i + len(x)].data)
            i += len(x)

    @property
    def n_par(self):
        r"""Returns the total number of trainable parameters in the machine."""
        return self._n_par

    def log_val(self, x, out=None):
        if len(x.shape) == 1:
            x = x[_np.newaxis, :]
        batch_shape = x.shape[:-1]

        with _torch.no_grad():
            t_out = self._module(_torch.from_numpy(x)).numpy().view(_np.complex128)
        if out is None:
            return t_out.reshape(batch_shape)
        _np.copyto(out, t_out.reshape(-1))

        return out

    def _der_log_backpack(self, x, out):
        def jacobian():
            return _torch.cat(
                [
                    p.grad_batch.flatten(start_dim=1)
                    for p in self._module.parameters()
                    if p.requires_grad
                ],
                dim=1,
            ).numpy()

        m_sum_real = self._module(x)[:, 0].sum(dim=0)
        with backpack.backpack(backpack.extensions.BatchGrad()):
            m_sum_real.backward()
        out.real = jacobian()

        m_sum_imag = self._module(x)[:, 1].sum(dim=0)
        with backpack.backpack(backpack.extensions.BatchGrad()):
            m_sum_imag.backward()
        out.imag = jacobian()

        self._module.zero_grad()
        return out

    def der_log(self, x, out=None):
        x = x.reshape(-1, x.shape[-1])
        if out is None:
            out = _np.empty([x.shape[0], self._n_par], dtype=_np.complex128)
        x = _torch.tensor(x, dtype=self._dtype)

        if self._use_backpack:
            return self._der_log_backpack(x, out)
        else:
            m = self._module(x)
            for i in range(x.size(0)):
                dws_real = _torch.autograd.grad(
                    m[i, 0], self._parameters, retain_graph=True
                )
                dws_imag = _torch.autograd.grad(
                    m[i, 1], self._parameters, retain_graph=True
                )
                out[i, ...].real = _torch.cat([dw.flatten() for dw in dws_real]).numpy()
                out[i, ...].imag = _torch.cat([dw.flatten() for dw in dws_imag]).numpy()
            self._module.zero_grad()
            return out

    def vector_jacobian_prod(
        self, x, vec, out=None, conjugate=True, return_jacobian=False
    ):
        x = x.reshape(-1, x.shape[-1])
        vec = vec.flatten()
        if out is None:
            out = _np.empty(self._n_par, dtype=_np.complex128)

        if return_jacobian:
            jac = self.der_log(x)
            out[...] = _np.einsum("n,nk->k", vec, jac.conj())
            if not conjugate:
                out[...] = out.conj()
            return out, jac
        else:
            x = _torch.tensor(x, dtype=self._dtype)
            y = self._module(x)
            self._module.zero_grad()
            vecj = _torch.empty(x.shape[0], 2, dtype=_torch.float64)

            def gradient():
                return _torch.cat(
                    [
                        p.grad.flatten()
                        for p in self._module.parameters()
                        if p.requires_grad
                    ]
                ).numpy()

            vecj[:, 0] = _torch.from_numpy(vec.real)
            vecj[:, 1] = _torch.from_numpy(vec.imag)
            y.backward(vecj, retain_graph=True)
            out.real = gradient()
            self._module.zero_grad()

            vecj[:, 0] = _torch.from_numpy(vec.imag)
            vecj[:, 1] = _torch.from_numpy(-vec.real)
            if not conjugate:
                vecj *= -1.0
            y.backward(vecj)
            out.imag = gradient()
            self._module.zero_grad()
            return out

    @property
    def state_dict(self):
        from collections import OrderedDict

        return OrderedDict(
            [(k, v.detach().numpy()) for k, v in self._module.state_dict().items()]
        )

    def __repr__(self):
        head = super().__repr__()
        module = "\n│  ".join([""] + repr(self._module).split("\n"))
        return head + module + "\n└"


class TorchLogCosh(_torch.nn.Module):
    """
    Log(cosh) activation function for PyTorch modules
    """

    def __init__(self):
        """
        Init method.
        """
        super().__init__()  # init the base class

    def forward(self, input):
        """
        Forward pass of the function.
        """
        return -input + _torch.nn.functional.softplus(2.0 * input)


class TorchView(_torch.nn.Module):
    """
    Reshaping layer for PyTorch modules
    """

    def __init__(self, shape):
        """
        Init method.
        """
        super().__init__()  # init the base class
        self.shape = shape

    def forward(self, x):
        return x.view(*self.shape)



from pytest import approx, raises
import numpy as np

import netket.legacy as nk
import netket.legacy.variational as vmc

from io import StringIO
from contextlib import redirect_stderr
import tempfile
import re

import pytest

pytestmark = pytest.mark.legacy

SEED = 214748364
nk.random.seed(SEED)


def _setup_vmc(lsq_solver=None):
    L = 4
    g = nk.graph.Hypercube(length=L, n_dim=1)
    hi = nk.hilbert.Spin(s=0.5, N=g.n_nodes)

    ma = nk.machine.RbmSpin(hilbert=hi, alpha=1)
    ma.init_random_parameters(sigma=0.01)

    ha = nk.operator.Ising(hi, graph=g, h=1.0)
    sa = nk.sampler.ExactSampler(machine=ma, sample_size=16)
    op = nk.optimizer.Sgd(ma, learning_rate=0.05)

    # Add custom observable
    X = [[0, 1], [1, 0]]
    sx = nk.operator.LocalOperator(hi, [X] * L, [[i] for i in range(8)])

    sr = nk.optimizer.SR(ma, use_iterative=False, lsq_solver=lsq_solver)
    driver = nk.Vmc(ha, sa, op, 1000, sr=sr)

    return ha, sx, ma, sa, driver


def test_before_first_step():
    ha, *_, driver = _setup_vmc()
    with raises(RuntimeError):
        doesnotwork = driver.estimate(ha)
    driver.advance(1)
    driver.estimate(ha)


def test_vmc_functions():
    ha, sx, ma, sampler, driver = _setup_vmc()
    driver.advance(500)

    state = ma.to_array()

    exact_dist = np.abs(state) ** 2

    n_samples = 16000

    for op, name in (ha, "ha"), (sx, "sx"):
        print("Testing expectation of op={}".format(name))

        states = np.array(list(ma.hilbert.states()))
        exact_locs = nk.operator.local_values(op, ma, states)
        exact_ex = np.sum(exact_dist * exact_locs).real

        for _ in sampler.samples(10):
            pass
        samples = sampler.generate_samples(n_samples).reshape((-1, ma.hilbert.size))

        der_logs = ma.der_log(samples)
        der_logs -= nk.stats.mean(der_logs)

        local_values = nk.operator.local_values(op, ma, samples)

        mean = local_values.mean()
        var = local_values.var()
        print(mean, var)

        # 5-sigma test for expectation values
        tol = np.sqrt(var / float(local_values.size)) * 5
        assert mean.real == approx(exact_ex.real, abs=tol)

        stats = vmc.estimate_expectations(
            op, sampler, n_samples=n_samples, n_discard=10
        )

        assert stats.mean.real == approx(mean.real, rel=tol)
        assert stats.mean.real == approx(exact_ex.real, abs=tol)

    local_values = nk.operator.local_values(ha, ma, samples)
    # grad = nk.stats.covariance_sv(local_values, der_logs)
    # assert grad.shape == (ma.n_par,)
    # assert np.mean(np.abs(grad) ** 2) == approx(0.0, abs=1e-8)

    _, grads = vmc.estimate_expectations(
        ha, sampler, n_samples, 10, compute_gradients=True
    )

    assert grads.shape == (ma.n_par,)
    assert np.mean(np.abs(grads) ** 2) == approx(0.0, abs=1e-8)


def test_raise_n_iter():
    ha, sx, ma, sampler, driver = _setup_vmc()
    with raises(
        ValueError,
    ):
        driver.run("prova", 12)


def test_vmc_use_cholesky_compatibility():
    ha, _, ma, sampler, _ = _setup_vmc(lsq_solver="Cholesky")


def test_vmc_progress_bar():
    ha, sx, ma, sampler, driver = _setup_vmc()
    tempdir = tempfile.mkdtemp()
    prefix = tempdir + "/vmc_progressbar_test"

    f = StringIO()
    with redirect_stderr(f):
        driver.run(5, prefix)
    pbar = f.getvalue().split("\r")[-1]
    assert re.match(r"100%\|#*\| (\d+)/\1", pbar)
    assert re.search(r"Energy=[-+]?[0-9]*\.?[0-9]*", pbar)
    assert re.search(r"σ²=[-+]?[0-9]*\.?[0-9]*", pbar)

    f = StringIO()
    with redirect_stderr(f):
        driver.run(5, prefix, show_progress=None)
    pbar = f.getvalue()
    assert not len(pbar)


def _energy(par, machine, H):
    machine.parameters = np.copy(par)
    psi = machine.to_array()
    return np.real(psi.conj() @ H @ psi)


def central_diff_grad(func, x, eps, *args):
    grad = np.zeros(len(x), dtype=complex)
    epsd = np.zeros(len(x), dtype=complex)
    epsd[0] = eps
    for i in range(len(x)):
        assert not np.any(np.isnan(x + epsd))
        grad_r = 0.5 * (func(x + epsd, *args) - func(x - epsd, *args))
        grad_i = 0.5 * (func(x + 1j * epsd, *args) - func(x - 1j * epsd, *args))
        grad[i] = 0.5 * grad_r + 0.5j * grad_i
        assert not np.isnan(grad[i])
        grad[i] /= eps
        epsd = np.roll(epsd, 1)
    return grad


def same_derivatives(der_log, num_der_log, abs_eps=1.0e-6, rel_eps=1.0e-6):
    assert np.max(np.real(der_log - num_der_log)) == approx(
        0.0, rel=rel_eps, abs=abs_eps
    )
    # The imaginary part is a bit more tricky, there might be an arbitrary phase shift
    assert np.max(
        np.abs(np.exp(np.imag(der_log - num_der_log) * 1.0j) - 1.0)
    ) == approx(0.0, rel=rel_eps, abs=abs_eps)


def test_vmc_gradient():
    ha, sx, ma, sampler, driver = _setup_vmc()
    pars = np.copy(ma.parameters)
    driver._sr = None
    grad_exact = central_diff_grad(_energy, pars, 1.0e-5, ma, ha.to_sparse())

    driver.n_samples = 1e6
    driver.n_discard = 1e3
    driver.machine.parameters = np.copy(pars)
    grad_approx = driver._forward_and_backward()

    err = 6 / np.sqrt(driver.n_samples)
    same_derivatives(grad_approx, grad_exact, abs_eps=err, rel_eps=1.0e-3)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from typing import Tuple, Optional, Union, Iterable

import jax.numpy as jnp
import numpy as np

from netket.utils import warn_deprecation

max_states = np.iinfo(np.int32).max
"""int: Maximum number of states that can be indexed"""


class NoneTypeT:
    pass


NoneType = NoneTypeT()


legacy_warn_str = (
    "This choice of arguments for `hilbert.random_state` is deprecated and "
    "will be removed in a future version.\n"
    "The new syntax is\n"
    "    hilbert.random_state(prngkey, size=1, dtype=jnp.float32)\n"
    "and, like in `jax.random`, the first argument is mandatory and must be a valid jax `PRNGKey`."
    "Results may differ between the states generated by the legacy, "
    "deprecated code and the new one. Note that the old code does not "
    "support defining custom methods."
)


class AbstractHilbert(abc.ABC):
    """Abstract class for NetKet hilbert objects.

    This class definese the common interface that can be used to
    interact with hilbert spaces.

    Hilbert Spaces are immutable.
    """

    def __init__(self):
        self._hash = None

    @property
    @abc.abstractmethod
    def size(self) -> int:
        r"""The number number of degrees of freedom in the basis of this
        Hilbert space."""
        raise NotImplementedError()  # pragma: no cover

    # after removing legacy:
    # signature must be the following
    # def random_state(self, key, size=None, dtype=np.float32):
    def random_state(
        self,
        key=NoneType,
        size: Optional[int] = NoneType,
        dtype=np.float32,
        out: Optional[np.ndarray] = None,
        rgen=None,
    ) -> jnp.ndarray:
        r"""Generates either a single or a batch of uniformly distributed random states.
        Runs as :code:`random_state(self, key, size=None, dtype=np.float32)` by default.

        Args:
            key: rng state from a jax-style functional generator.
            size: If provided, returns a batch of configurations of the form
                  :code:`(size, N)` if size is an integer or :code:`(*size, N)` if it is
                  a tuple and where :math:`N` is the Hilbert space size.
                  By default, a single random configuration with shape
                  :code:`(#,)` is returned.
            dtype: DType of the resulting vector.
            out: Deprecated. Will be removed in v3.1
            rgen: Deprecated. Will be removed in v3.1

        Returns:
            A state or batch of states sampled from the uniform distribution on the
            hilbert space.

        Example:

            >>> import netket, jax
            >>> hi = netket.hilbert.Qubit(N=2)
            >>> k1, k2 = jax.random.split(jax.random.PRNGKey(1))
            >>> print(hi.random_state(key=k1))
            [1. 0.]
            >>> print(hi.random_state(key=k2, size=2))
            [[0. 0.]
             [0. 1.]]
        """
        # legacy support
        # TODO: Remove in 3.1
        # if no positional arguments, and key is unspecified -> legacy
        if key is NoneType:
            warn_deprecation(legacy_warn_str)
            # legacy sure
            if size is NoneType:
                return self._random_state_legacy(size=None, out=out, rgen=rgen)
            else:
                return self._random_state_legacy(size=size, out=out, rgen=rgen)
        elif isinstance(key, tuple) or isinstance(key, int) and size is NoneType:
            # if one positional argument legacy typee...
            warn_deprecation(legacy_warn_str)
            return self._random_state_legacy(size=key, out=out, rgen=rgen)
        else:
            from netket.hilbert import random

            size = size if size is not NoneType else None

            return random.random_state(self, key, size, dtype=dtype)

    def ptrace(self, sites: Union[int, Iterable]) -> "AbstractHilbert":
        """Returns the hilbert space without the selected sites.

        Not all hilbert spaces support this operation.

        Args:
            sites: a site or list of sites to trace away

        Returns:
            The partially-traced hilbert space. The type of the resulting hilbert space
            might be different from the starting one.
        """
        raise NotImplementedError("Ptrace not implemented for this hilbert space type.")

    @property
    def is_indexable(self) -> bool:
        """Whever the space can be indexed with an integer"""
        return False

    @property
    @abc.abstractmethod
    def _attrs(self) -> Tuple:
        """
        Tuple of hashable attributs, used to compute the immutable
        hash of this Hilbert space
        """

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self._attrs == other._attrs

        return False

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self._attrs)

        return self._hash



from ..abstract_optimizer import AbstractOptimizer
import numpy as _np


class AdaGrad(AbstractOptimizer):
    r"""AdaGrad Optimizer.
        In many cases, in Sgd the learning rate :math`\eta` should
        decay as a function of training iteration to prevent overshooting
        as the optimum is approached. AdaGrad is an adaptive learning
        rate algorithm that automatically scales the learning rate with a sum
        over past gradients. The vector :math:`\mathbf{g}` is initialized to zero.
        Given a stochastic estimate of the gradient of the cost function :math:`G(\mathbf{p})`,
        the updates for :math:`g_k` and the parameter :math:`p_k` are


        .. math:: g^\prime_k &= g_k + G_k(\mathbf{p})^2\\
                  p^\prime_k &= p_k - \frac{\eta}{\sqrt{g_k + \epsilon}}G_k(\mathbf{p})

        AdaGrad has been shown to perform particularly well when
        the gradients are sparse, but the learning rate may become too small
        after many updates because the sum over the squares of past gradients is cumulative.
    """

    def __init__(self, learning_rate=0.001, epscut=1.0e-7):
        r"""
        Constructs a new ``AdaGrad`` optimizer.

        Args:
            learning_rate: Learning rate :math:`\eta`.
            epscut: Small :math:`\epsilon` cutoff.

        Examples:
            Simple AdaGrad optimizer.

            >>> from netket.optimizer import AdaGrad
            >>> op = AdaGrad()
        """

        self._eta = learning_rate
        self._epscut = epscut

        if epscut <= 0:
            raise ValueError("Invalid epsilon cutoff.")
        if learning_rate < 0:
            raise ValueError("Invalid learning rate.")

        self._Gt = None

    def update(self, grad, pars):
        if self._Gt is None:
            self._Gt = _np.zeros(pars.shape[0])

        self._Gt += _np.abs(grad) ** 2

        pars -= self._eta * grad / _np.sqrt(self._Gt + self._epscut)

        return pars

    def reset(self):
        if self._mt is not None:
            self._Gt.fill(0.0)

    def __repr__(self):
        rep = "AdaGrad optimizer with these parameters :"
        rep += "\nLearning Rate = " + str(self._eta)
        rep += "\nepscut = " + str(self._epscut)
        return rep



import json as _json
from os import path as _path
from netket.legacy.vmc_common import tree_map as _tree_map
from netket.legacy.stats.mc_stats import Stats

from tensorboardX import SummaryWriter


def tree_log(tree, root, data):
    """
    Maps all elements in tree, recursively calling tree_log with a new root string,
    and when it reaches leaves pushes (string, leave) tuples to data.
    """
    if tree is None:
        return data
    elif isinstance(tree, list):
        tmp = [
            tree_log(val, root + "/{}".format(i), data) for (i, val) in enumerate(tree)
        ]
        return data
    elif isinstance(tree, list) and hasattr(tree, "_fields"):
        tmp = [
            tree_log(getattr(tree, key), root + "/{}".format(key), data)
            for key in tree._fields
        ]
        return data
    elif isinstance(tree, tuple):
        tmp = tuple(
            tree_log(val, root + "/{}".format(i), data) for (i, val) in enumerate(tree)
        )
        return data
    elif isinstance(tree, dict):
        return {
            key: tree_log(value, root + "/{}".format(key), data)
            for key, value in tree.items()
        }
    else:
        data.append((root, tree))
        return data


class TBLog:
    """
    Creates a tensorboard logger using tensorboardX's summarywriter.
    Refer to its documentation for further details

    https://tensorboardx.readthedocs.io/en/latest/tensorboard.html

    Args:
        logdir (string): Save directory location. Default is
          runs/**CURRENT_DATETIME_HOSTNAME**, which changes after each run.
          Use hierarchical folder structure to compare
          between runs easily. e.g. pass in 'runs/exp1', 'runs/exp2', etc.
          for each new experiment to compare across them.
        comment (string): Comment logdir suffix appended to the default
          ``logdir``. If ``logdir`` is assigned, this argument has no effect.
        purge_step (int):
          When logging crashes at step :math:`T+X` and restarts at step :math:`T`,
          any events whose global_step larger or equal to :math:`T` will be
          purged and hidden from TensorBoard.
          Note that crashed and resumed experiments should have the same ``logdir``.
        max_queue (int): Size of the queue for pending events and
          summaries before one of the 'add' calls forces a flush to disk.
          Default is ten items.
        flush_secs (int): How often, in seconds, to flush the
          pending events and summaries to disk. Default is every two minutes.
        filename_suffix (string): Suffix added to all event filenames in
          the logdir directory. More details on filename construction in
          tensorboard.summary.writer.event_file_writer.EventFileWriter.
        write_to_disk (boolean):
          If pass `False`, TBLog will not write to disk.
    Examples::
        import netket as nk
        # create a summary writer with automatically generated folder name.
        writer = nk.logging.TBLog()
        # folder location: runs/May04_22-14-54_s-MacBook-Pro.local/
        # create a summary writer using the specified folder name.
        writer = nk.logging.TBLog("my_experiment")
        # folder location: my_experiment
        # create a summary writer with comment appended.
        writer = nk.logging.TBLog(comment="LR_0.1_BATCH_16")
        # folder location: runs/May04_22-14-54_s-MacBook-Pro.localLR_0.1_BATCH_16/
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):

        self._writer = SummaryWriter(*args, **kwargs)

        self._old_step = 0

    def __call__(self, step, item, machine):

        data = []
        tree_log(item, "", data)

        for key, val in data:
            if isinstance(val, Stats):
                val = val.mean

            if isinstance(val, complex):
                self._writer.add_scalar(key[1:] + "/re", val.real, step)
                self._writer.add_scalar(key[1:] + "/im", val.imag, step)
            else:
                self._writer.add_scalar(key[1:], val, step)

        self._writer.flush()
        self._old_step = step

    def _flush_log(self):
        self._writer.flush()

    def _flush_params(self, machine):
        return None

    def flush(self, machine=None):
        """
        Writes to file the content of this logger.

        :param machine: optionally also writes the parameters of the machine.
        """
        self._flush_log()

        if machine is not None:
            self._flush_params(machine)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import netket as nk
import numpy as np
import pytest
from jax import numpy as jnp


@pytest.mark.parametrize("dtype", [jnp.float64, jnp.complex128])
@pytest.mark.parametrize(
    "hilbert",
    [
        pytest.param(
            nk.hilbert.Spin(s=1 / 2, N=4),
            id="spin_1/2",
        ),
        pytest.param(
            nk.hilbert.Spin(s=1, N=4),
            id="spin_1",
        ),
        pytest.param(
            nk.hilbert.Fock(n_max=3, N=4),
            id="fock",
        ),
    ],
)
@pytest.mark.parametrize(
    "partial_model",
    [
        pytest.param(
            lambda hilbert, dtype: nk.models.ARNNDense(
                hilbert=hilbert,
                layers=3,
                features=5,
                dtype=dtype,
            ),
            id="dense",
        ),
        pytest.param(
            lambda hilbert, dtype: nk.models.ARNNConv1D(
                hilbert=hilbert,
                layers=3,
                features=5,
                kernel_size=2,
                dtype=dtype,
            ),
            id="conv1d",
        ),
        pytest.param(
            lambda hilbert, dtype: nk.models.ARNNConv1D(
                hilbert=hilbert,
                layers=3,
                features=5,
                kernel_size=2,
                kernel_dilation=2,
                dtype=dtype,
            ),
            id="conv1d_dilation",
        ),
    ],
)
def test_ARNN(partial_model, hilbert, dtype):
    batch_size = 3

    model = partial_model(hilbert, dtype)

    key_spins, key_model = jax.random.split(jax.random.PRNGKey(0))
    spins = hilbert.random_state(key_spins, size=batch_size)
    (p, _), params = model.init_with_output(
        key_model, spins, None, method=model.conditionals
    )

    # Test if the model is normalized
    # The result may not be very accurate, because it is in exp space
    psi = nk.nn.to_array(hilbert, model.apply, params, normalize=False)
    assert psi.conj() @ psi == pytest.approx(1, rel=1e-5, abs=1e-5)

    # Test if the model is autoregressive
    for i in range(batch_size):
        for j in range(hilbert.size):
            # Change one input element at a time
            spins_new = spins.at[i, j].set(-spins[i, j])
            p_new, _ = model.apply(params, spins_new, None, method=model.conditionals)
            p_diff = p_new - p

            # The former output elements should not change
            p_diff = p_diff.at[i, j + 1 :].set(0)

            np.testing.assert_allclose(p_diff, 0, err_msg=f"i={i} j={j}")


def test_throwing():
    def build_model(hilbert):
        nk.models.ARNNConv1D(hilbert=hilbert, layers=3, features=5, kernel_size=2)

    # Only homogeneous Hilbert spaces are supported
    with pytest.raises(ValueError):
        hilbert = nk.hilbert.Spin(s=1 / 2, N=4)
        hilbert = nk.hilbert.DoubledHilbert(hilbert)
        build_model(None)

    # Only unconstrained Hilbert spaces are supported
    with pytest.raises(ValueError):
        hilbert = nk.hilbert.Fock(n_max=3, N=4, n_particles=3)
        build_model(hilbert)



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax.experimental import stax

from collections import OrderedDict
import warnings
import igraph

from .abstract_machine import AbstractMachine

import numpy as _np
from jax import numpy as jnp
from jax import random
from netket.graph import Chain
from netket.legacy.random import randint as _randint
from netket import utils
from jax.tree_util import tree_flatten, tree_unflatten, tree_map, tree_leaves

from ._jax_utils import forward_apply, tree_size, grad as nk_grad, vjp as nk_vjp

# An extensible list of wrapping package modules like flax, haiku and others
jax_package_wrappers = []


def add_package_wrapper(condition_fun, wrapper_fun):
    """
    add_package_wrapper(condition_fun, wrapper_fun)

    Add support for a jax-based package.
    `condition_fun(module)` must return true if `module` is a module of this package,
    and `wrapper_fun(module)` must return the tuple of init and apply functions for
    this module.
    """
    jax_package_wrappers.append((condition_fun, wrapper_fun))


# Maybe wrap the module of some package.
def maybe_wrap_module(module):
    for (condition_fun, wrap_fun) in jax_package_wrappers:
        if condition_fun(module):
            return wrap_fun(module)

    return module


class Jax(AbstractMachine):
    def __init__(self, hilbert, module, dtype=complex):
        """
        Wraps a stax network (which is a tuple of `init_fn` and `predict_fn`)
        so that it can be used as a NetKet machine.

        Args:
            hilbert: Hilbert space on which the state is defined. Should be a
                subclass of `netket.hilbert.Hilbert`.
            module: A pair `(init_fn, predict_fn)`. See the documentation of
                jax.experimental.stax` for more info.
            dtype: either complex or float, is the type used for the weights.
                In both cases the network must have a single output.
        """
        super().__init__(hilbert=hilbert, dtype=dtype)

        self._init_fn, self._forward_fn_nj = maybe_wrap_module(module)

        self._forward_fn = lambda pars, x: forward_apply(pars, self._forward_fn_nj, x)

        self.jax_init_parameters()

        self._npar = tree_size(self._params)

    def jax_init_parameters(self, seed=None):
        """
        Uses the init function of the jax networks to generate a set of parameters.
        Beware that usually jax does not correctly set the imaginary part of
        networks, so for complex networks it MUST be followed by a call to
        init_random_parameters, unless your network correctly initialised the
        imaginary part.
        """
        if seed is None:
            seed = _randint(0, 2 ** 32 - 2, size=())

        input_shape = (-1, self.input_size)
        output_shape, params = self._init_fn(jax.random.PRNGKey(seed), input_shape)

        self._params = self._cast(params)

        # TODO
        # We should also test that the output_shape is
        # (1,) or () for a vector input_shape (N,), which is what
        # is provided by the sampler.
        #  Either add this check of fix the samplers
        if output_shape != (-1, 1) and output_shape != (-1,):
            raise ValueError(
                r"""
                A valid network must have 1 output, but the network
                you attempted to build, provided input_shape={} gave
                output_shape={}""".format(
                    input_shape, output_shape
                )
            )

    def init_random_parameters(self, seed=None, sigma=0.01):
        rgen = _np.random.RandomState(seed)

        pars = rgen.normal(scale=sigma, size=self.n_par)

        if self.has_complex_parameters:
            pars = pars + 1.0j * rgen.normal(scale=sigma, size=self.n_par)

        self.parameters = self.numpy_unflatten(pars, self.parameters)

    def _cast(self, p):
        if self._dtype is complex:
            from jax.tree_util import tree_unflatten, tree_flatten

            # TODO use tree_map instead
            value_flat, value_tree = tree_flatten(p)
            values_c = list(map(lambda v: v.astype(jax.numpy.complex128), value_flat))

            return tree_unflatten(value_tree, values_c)
        else:
            return p

    @property
    def n_par(self):
        r"""The number of variational parameters in the machine."""
        return self._npar

    def log_val(self, x, out=None):
        if x.ndim != 2:
            raise RuntimeError("Invalid input shape, expected a 2d array")

        if out is None:
            out = forward_apply(self._params, self._forward_fn_nj, x).reshape(
                x.shape[0],
            )
        else:
            out[:] = forward_apply(self._params, self._forward_fn_nj, x).reshape(
                x.shape[0],
            )
        return out

    @property
    def jax_forward(self):
        return self._forward_fn

    def der_log(self, x, out=None):
        if x.ndim != 2:
            raise RuntimeError("Invalid input shape, expected a 2d array")

        # Jax has bugs for R->C functions...
        out = nk_grad(self._params, self._forward_fn_nj, x)

        return out

    def vector_jacobian_prod(
        self, x, vec, out=None, conjugate=True, return_jacobian=False
    ):
        r"""Computes the scalar product between gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and a vector `vec`. The result is stored into `out`.

        Args:
             x: a matrix of `float64` of shape `(*, self.n_visible)`.
             vec: a `complex128` vector used to compute the inner product with the jacobian.
             out: The result of the inner product, it is a vector of `complex128` and length `self.n_par`.
             conjugate (bool): If true, this computes the conjugate of the vector jacobian product.
             return_jacobian (bool): If true, the Jacobian is explicitely computed and returned.


        Returns:
             `out` only or (out,jacobian) if return_jacobian is True
        """
        if not return_jacobian:
            return nk_vjp(self._params, self._forward_fn_nj, x, vec, conjugate)

        else:

            if conjugate and self._dtype is complex:

                def prodj(j):
                    return jax.numpy.tensordot(vec.transpose(), j.conjugate(), axes=1)

            else:

                def prodj(j):
                    return jax.numpy.tensordot(vec.transpose().conjugate(), j, axes=1)

            jacobian = nk_grad(self._params, self._forward_fn_nj, x)
            out = tree_map(prodj, jacobian)

            return out, jacobian

    @property
    def state_dict(self):
        state = []
        for i, layer in enumerate(self._params):
            for j, p in enumerate(layer):
                state.append((str((i, j)), p))
        return OrderedDict(state)

    @property
    def parameters(self):
        return self._params

    @parameters.setter
    def parameters(self, p):
        assert tree_size(p) == self._npar
        self._params = p

    def numpy_flatten(self, data):
        r"""Returns a flattened numpy array representing the given data.
            This is typically used to serialize parameters and gradients.

        Args:
             data: a (possibly non-flat) structure containing jax arrays.

        Returns:
             numpy.ndarray: a one-dimensional array containing a copy of data
        """

        return _np.concatenate(tuple(fd.reshape(-1) for fd in tree_leaves(data)))

    def numpy_unflatten(self, data, shape_like):
        r"""Attempts a deserialization of the given numpy data.
            This is typically used to deserialize parameters and gradients.

        Args:
             data: a 1d numpy array.
             shape_like: this as in instance having the same type and shape of
                         the desired conversion.

        Returns:
             A possibly non-flat structure of jax arrays containing a copy of data
             compatible with the given shape.
        """
        shf, tree = tree_flatten(shape_like)

        datalist = []
        k = 0
        for s in shf:
            size = s.size
            datalist.append(jax.numpy.asarray(data[k : k + size]).reshape(s.shape))
            k += size

        return tree_unflatten(tree, datalist)


from jax.experimental import stax
from jax.experimental.stax import Dense
from jax.nn.initializers import glorot_normal, normal


def DenseReal(out_dim, use_bias=True, W_init=glorot_normal(), b_init=normal()):
    """Layer constructor function for a dense (fully-connected) layer with complex output but real weight."""

    def init_fun(rng, input_shape):
        output_shape = input_shape[:-1] + (out_dim,)
        k1, k2 = random.split(rng)
        if use_bias:
            Wr, Wi, br, bi = (
                W_init(k1, (input_shape[-1], out_dim)),
                W_init(k1, (input_shape[-1], out_dim)),
                b_init(k2, (out_dim,)),
                b_init(k2, (out_dim,)),
            )
            return output_shape, (Wr, Wi, br, bi)
        else:
            Wr, Wi = (
                W_init(k1, (input_shape[-1], out_dim)),
                W_init(k1, (input_shape[-1], out_dim)),
            )
            return output_shape, (Wr, Wi)

    def apply_fun(params, inputs, **kwargs):
        # distinguish between real and complex input because
        # jax is not smart enough to use the good gemm on his own.
        if jnp.iscomplexobj(inputs):
            if use_bias:
                Wr, Wi, br, bi = params
                return jnp.dot(inputs, Wr + 1j * Wi) + br + 1j * bi
            else:
                Wr, Wi = params
                return jnp.dot(inputs, Wr + 1j * Wi)
        else:
            if use_bias:
                Wr, Wi, br, bi = params
                return jnp.dot(inputs, Wr) + 1j * jnp.dot(inputs, Wi) + br + 1j * bi
            else:
                Wr, Wi = params
                return jnp.dot(inputs, Wr) + 1j * jnp.dot(inputs, Wi)

    return init_fun, apply_fun


def SumLayer():
    def init_fun(rng, input_shape):
        output_shape = (-1, 1)
        return output_shape, ()

    def apply_fun(params, inputs, **kwargs):
        return inputs.sum(axis=-1)

    return init_fun, apply_fun


SumLayer = SumLayer()


def logcosh(x):
    x = x * jax.numpy.sign(x.real)
    return x + jax.numpy.log(1.0 + jax.numpy.exp(-2.0 * x)) - jax.numpy.log(2.0)


LogCoshLayer = stax.elementwise(logcosh)


def JaxRbm(hilbert, alpha, dtype=complex):
    return Jax(
        hilbert,
        stax.serial(stax.Dense(alpha * hilbert.size), LogCoshLayer, SumLayer),
        dtype=dtype,
    )


def MPSPeriodic(hilbert, graph, bond_dim, diag=False, symperiod=None, dtype=complex):
    r"""
    Constructs a periodic Matrix Product State (MPS) for a quantum state of discrete
    degrees of freedom, wrapped as Jax machine.  The MPS is defined as

    .. math:: \Psi(s_1,\dots s_N) = \Tr(A[s_1]\dots A[s_N]),

    for arbitrary local quantum numbers :math:`s_i`, where :math:`A[s_1]` is a matrix
    of dimension (bdim,bdim), depending on the value of the local quantum number :math:`s_i`.

        Args:
            hilbert: Hilbert space on which the state is defined. Should be a
                subclass of `netket.hilbert.Hilbert`.
            bond_dim (int): Virtual dimension of the MPS tensors.
            diag (bool): Whether or not to use diagonal matrices in the MPS tensors.
                default=False
            symperiod (int): Periodicity in the chain of MPS tensors. The chain of
                MPS tensors is constructed as a sequence of identical unit cells
                consisting of symperiod tensors. if None, symperiod equals the
                number of physical degrees of freedom.
                default=None
            dtype: complex or float, whether the variational parameters of the MPS
                are real or complex. default=complex

        returns:
            Jax machine of the Matrix Product state.
    """
    return Jax(
        hilbert,
        mpsPeriodic(hilbert, graph, bond_dim, diag, symperiod, dtype),
        dtype=dtype,
    )


def mpsPeriodic(hilbert, graph, bond_dim, diag=False, symperiod=None, dtype=complex):
    # default standard deviation equals 1e-2
    normal_init = jax.nn.initializers.normal()

    L = hilbert.size
    phys_dim = hilbert.local_size
    diag = diag
    dtype = dtype

    # determine transformation from local states to indices
    local_states = jax.numpy.array(hilbert.local_states)
    loc_vals_spacing = jax.numpy.roll(local_states, -1)[0:-1] - local_states[0:-1]
    if jax.numpy.max(loc_vals_spacing) == jax.numpy.min(loc_vals_spacing):
        loc_vals_spacing = loc_vals_spacing[0]
    else:
        raise AssertionError(
            "JaxMpsPeriodic can only be used with evenly spaced hilbert local values"
        )
    loc_vals_bias = jax.numpy.min(local_states)

    # check whether graph is periodic chain
    chain_graph = Chain(graph.n_edges).to_igraph()
    if not igraph.Graph(edges=graph.edges()).isomorphic(chain_graph):
        warnings.warn(
            "Warning: graph is not isomorphic to chain with periodic boundary conditions",
            UserWarning,
        )

    # determine shape of unit cell
    if symperiod == None:
        symperiod = L
    if L % symperiod == 0 and symperiod > 0:
        if diag:
            unit_cell_shape = (symperiod, phys_dim, bond_dim)
        else:
            unit_cell_shape = (symperiod, phys_dim, bond_dim, bond_dim)
    else:
        raise AssertionError(
            "The number of degrees of freedom of the Hilbert space needs to be a multiple of the period of the MPS"
        )

    # define diagonal tensors with correct unit cell shape
    if diag:
        iden_tensors = jax.numpy.ones((symperiod, phys_dim, bond_dim), dtype=dtype)
    else:
        iden_tensors = jax.numpy.repeat(
            jax.numpy.eye(bond_dim, dtype=dtype)[jax.numpy.newaxis, :, :],
            symperiod * phys_dim,
            axis=0,
        )
        iden_tensors = iden_tensors.reshape(symperiod, phys_dim, bond_dim, bond_dim)

    def init_fun(rng, input_shape):
        rng_re, rng_im = jax.random.split(rng)
        random_tensors_real = normal_init(rng_re, unit_cell_shape)
        if dtype == complex:
            random_tensors_imag = normal_init(rng_im, unit_cell_shape)
            random_tensors = random_tensors_real + 1j * random_tensors_imag
        else:
            random_tensors = random_tensors_real

        tensors = random_tensors + iden_tensors

        return (-1, 1), (tensors)

    def apply_fun(params, x, **kwargs):
        # expand diagonal to square matrices if diagonal mps
        if diag:
            params = jax.numpy.einsum(
                "ijk,kl->ijkl", params, jax.numpy.eye(params.shape[-1])
            )

        # create all tensors in mps from unit cell
        all_tensors = jax.numpy.tile(params, (L // symperiod, 1, 1, 1))

        # transform input to indices
        x = (x - loc_vals_bias) / loc_vals_spacing
        if len(x.shape) == 1:  # batch size is one
            x = jax.numpy.expand_dims(x, 0)

        def select_tensor(tensor, index):
            return tensor[index.astype(int)]

        def select_all_tensors(all_tensors, indices):
            return jax.vmap(select_tensor)(all_tensors, indices)

        # select right tensors using input for matrix multiplication
        selected_tensors = jax.vmap(select_all_tensors, (None, 0))(all_tensors, x)

        # create loop carry, in this case a unit matrix
        edges = jax.numpy.repeat(
            jax.numpy.eye(bond_dim, dtype=selected_tensors.dtype)[
                jax.numpy.newaxis, :, :
            ],
            selected_tensors.shape[0],
            axis=0,
        )

        def trace_mps(tensors, edge):
            def multiply_tensors(left_tensor, right_tensor):
                return jax.numpy.einsum("ij,jk->ik", left_tensor, right_tensor), None

            edge, _ = jax.lax.scan(multiply_tensors, edge, tensors)

            return jax.numpy.trace(edge)

        # trace the matrix multiplication
        z = jax.vmap(trace_mps)(selected_tensors, edges)
        return z

    return init_fun, apply_fun


def FanInSum2ModPhase():
    """Layer construction function for a fan-in sum layer."""

    def init_fun(rng, input_shape):
        output_shape = input_shape[0]
        return output_shape, tuple()

    def apply_fun(params, inputs, **kwargs):
        output = 1.0 * inputs[0] + 1.0j * inputs[1]
        return output

    return init_fun, apply_fun


FanInSum2ModPhase = FanInSum2ModPhase()


def jaxRbmSpinPhase(hilbert, alpha):
    return stax.serial(
        stax.FanOut(2),
        stax.parallel(
            stax.serial(stax.Dense(alpha * hilbert.size), LogCoshLayer, SumLayer),
            stax.serial(stax.Dense(alpha * hilbert.size), LogCoshLayer, SumLayer),
        ),
        FanInSum2ModPhase,
    )


def JaxRbmSpinPhase(hilbert, alpha, dtype=float):
    return Jax(
        hilbert,
        jaxRbmSpinPhase(hilbert, alpha),
        dtype=dtype,
    )



import jax
import numpy


class _JaxExchangeKernel:
    def __init__(self, hilbert, clusters):
        self.clusters = jax.numpy.asarray(clusters)

        self.clusters_size = self.clusters.shape[0]

        self._hilbert = hilbert

    def transition(self, key, state):

        # pick a random cluster
        cl = jax.random.randint(key, shape=(1,), minval=0, maxval=self.clusters_size)

        # sites to be exchanged
        si = self.clusters[cl, 0]
        sj = self.clusters[cl, 1]

        state_1 = jax.ops.index_update(state, si, state[sj])
        return jax.ops.index_update(state_1, sj, state[si])

    def random_state(self, key, state):
        return key, jax.numpy.asarray(self._hilbert.random_state())



import netket.legacy as nk
import networkx as nx
import numpy as np
import pytest
from pytest import approx
from scipy.stats import power_divergence, combine_pvalues, chisquare
from netket.legacy.random import randint

pytestmark = pytest.mark.legacy

samplers = {}

nk.random.seed(1234567)
np.random.seed(1234)

from netket.utils import jax_available as test_jax


# TESTS FOR SPIN HILBERT
# Constructing a 1d lattice
g = nk.graph.Hypercube(length=4, n_dim=1)

# Hilbert space of spins from given graph
hi = nk.hilbert.Spin(s=0.5, N=g.n_nodes)
ma = nk.machine.RbmSpin(hilbert=hi, alpha=1)
ma.init_random_parameters(sigma=0.2)

sa = nk.sampler.MetropolisLocal(machine=ma, n_chains=16)
samplers["MetropolisLocal RbmSpin"] = sa

hib = nk.hilbert.Boson(n_max=1, N=g.n_nodes, n_bosons=1)
mab = nk.machine.RbmSpin(hilbert=hib, alpha=1)
mab.init_random_parameters(sigma=0.2)
sa = nk.sampler.MetropolisExchange(machine=mab, n_chains=16, graph=g)
samplers["MetropolisExchange RbmSpin(boson)"] = sa

sa = nk.sampler.ExactSampler(machine=ma, sample_size=8)
samplers["Exact RbmSpin"] = sa

sa = nk.sampler.MetropolisLocalPt(machine=ma, n_replicas=4)
samplers["MetropolisLocalPt RbmSpin"] = sa

ha = nk.operator.Ising(hilbert=hi, graph=g, h=1.0)
sa = nk.sampler.MetropolisHamiltonian(machine=ma, hamiltonian=ha)
samplers["MetropolisHamiltonian RbmSpin"] = sa

# Test with uniform probability
maz = nk.machine.RbmSpin(hilbert=hi, alpha=1)
maz.init_random_parameters(sigma=0)
sa = nk.sampler.MetropolisLocal(machine=maz, sweep_size=hi.size + 1, n_chains=2)
samplers["MetropolisLocal RbmSpin ZeroPars"] = sa

mas = nk.machine.RbmSpinSymm(hilbert=hi, alpha=1, automorphisms=g)
mas.init_random_parameters(sigma=0.2)
sa = nk.sampler.MetropolisHamiltonianPt(machine=mas, hamiltonian=ha, n_replicas=4)
samplers["MetropolisHamiltonianPt RbmSpinSymm"] = sa

hi = nk.hilbert.Boson(N=g.n_nodes, n_max=3)
ma = nk.machine.RbmSpin(hilbert=hi, alpha=1)
ma.init_random_parameters(sigma=0.1)
sa = nk.sampler.MetropolisLocal(machine=ma)
samplers["MetropolisLocal Boson"] = sa

sa = nk.sampler.MetropolisLocalPt(machine=ma, n_replicas=2)
samplers["MetropolisLocalPt Boson"] = sa

hi = nk.hilbert.Boson(N=g.n_nodes, n_max=3)
ma = nk.machine.RbmSpin(hilbert=hi, alpha=1)
ma.init_random_parameters(sigma=0.1)
sa = nk.sampler.ExactSampler(machine=ma)
samplers["Exact Boson"] = sa

hi = nk.hilbert.Spin(s=0.5, N=g.n_nodes)
g = nk.graph.Hypercube(length=3, n_dim=1)
ma = nk.machine.RbmSpinSymm(hilbert=hi, alpha=1, automorphisms=g)
ma.init_random_parameters(sigma=0.2)
l = hi.size
X = [[0, 1], [1, 0]]

move_op = nk.operator.LocalOperator(
    hilbert=hi, operators=[X] * l, acting_on=[[i] for i in range(l)]
)


sa = nk.sampler.CustomSampler(machine=ma, move_operators=move_op)
samplers["CustomSampler Spin"] = sa


sa = nk.sampler.CustomSamplerPt(machine=ma, move_operators=move_op, n_replicas=4)
samplers["CustomSamplerPt Spin"] = sa

# Two types of custom moves
# single spin flips and nearest-neighbours exchanges
spsm = [[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]

ops = [X] * l
ops += [spsm] * l

acting_on = [[i] for i in range(l)]
acting_on += [[i, (i + 1) % l] for i in range(l)]

move_op = nk.operator.LocalOperator(hilbert=hi, operators=ops, acting_on=acting_on)

sa = nk.sampler.CustomSampler(machine=ma, move_operators=move_op)
samplers["CustomSampler Spin 2 moves"] = sa

# Diagonal density matrix sampling
ma = nk.machine.density_matrix.RbmSpin(
    hilbert=hi,
    alpha=1,
    use_visible_bias=True,
    use_hidden_bias=True,
)
ma.init_random_parameters(sigma=0.2)
dm = ma.diagonal()
sa = nk.sampler.MetropolisLocal(machine=dm)
samplers["Diagonal Density Matrix"] = sa

sa = nk.sampler.ExactSampler(machine=dm)
samplers["Exact Diagonal Density Matrix"] = sa

g = nk.graph.Hypercube(length=3, n_dim=1)
hi = nk.hilbert.Spin(s=0.5, N=g.n_nodes)
ma = nk.machine.density_matrix.RbmSpin(
    hilbert=hi,
    alpha=1,
    use_visible_bias=True,
    use_hidden_bias=True,
)

ma.init_random_parameters(sigma=0.2)
samplers["Metropolis Density Matrix"] = nk.sampler.MetropolisLocal(ma, n_chains=16)

sa = nk.sampler.ExactSampler(machine=ma, sample_size=8)
samplers["Exact Density Matrix"] = sa

if test_jax:
    ma = nk.machine.density_matrix.NdmSpinPhase(hilbert=hi, alpha=1, beta=1)
    ma.init_random_parameters(sigma=0.2)
    samplers["Metropolis Density Matrix Jax"] = nk.sampler.MetropolisLocal(
        ma, n_chains=16
    )

    ma = nk.machine.JaxRbm(hilbert=hi, alpha=1)
    ma.init_random_parameters(sigma=0.2)
    samplers["Metropolis Rbm Jax"] = nk.sampler.MetropolisLocal(ma, n_chains=16)

    hib = nk.hilbert.Boson(n_max=1, N=g.n_nodes, n_bosons=1)
    mab = nk.machine.JaxRbm(hilbert=hib, alpha=1)
    mab.init_random_parameters(sigma=0.2)
    sa = nk.sampler.MetropolisExchange(machine=mab, n_chains=16, graph=g)
    samplers["MetropolisExchange RbmSpin(boson) Jax"] = sa

    # Test a machine which only works with 2D output and not 1D
    import jax
    from jax.nn.initializers import glorot_normal

    def Jastrow(W_init=glorot_normal()):
        def init_fun(rng, input_shape):
            N = input_shape[-1]
            return input_shape[:-1], W_init(rng, (N, N))

        def apply_fun(W, x, **kwargs):
            return jax.vmap(
                lambda W, x: jax.numpy.einsum("i,ij,j", x, W, x), in_axes=(None, 0)
            )(W, x)

        return init_fun, apply_fun

    ma = nk.machine.Jax(hi, Jastrow(), dtype=float)
    ma.init_random_parameters(sigma=0.2)
    samplers["Metropolis Jastrow Jax"] = nk.sampler.MetropolisLocal(ma, n_chains=16)


def test_states_in_hilbert():
    for name, sa in samplers.items():
        print("Sampler test: %s" % name)

        ma = sa.machine
        hi = ma.hilbert
        localstates = hi.local_states

        for sample in sa.samples(100):
            assert sample.shape[1] == ma.input_size
            for v in sample.reshape(-1):
                assert v in localstates

        if hasattr(sa, "acceptance"):
            assert np.min(sa.acceptance) >= 0 and np.max(sa.acceptance) <= 1.0


# Testing that samples generated from direct sampling are compatible with those
# generated by markov chain sampling
# here we use a combination of power divergence tests


def test_correct_sampling():
    for name, sa in samplers.items():
        print("Sampler test: %s" % name)

        ma = sa.machine
        hi = ma.hilbert

        if ma.input_size == 2 * hi.size:
            hi = nk.hilbert.DoubledHilbert(hi)
        n_states = hi.n_states

        n_samples = max(40 * n_states, 10000)

        ord = randint(1, 3, size=()).item()
        assert ord == 1 or ord == 2

        sa.machine_pow = ord

        ps = np.absolute(ma.to_array()) ** ord
        ps /= ps.sum()

        n_rep = 6
        pvalues = np.zeros(n_rep)

        sa.reset(True)

        for jrep in range(n_rep):

            # Burnout phase
            samples = sa.generate_samples(n_samples // 10)

            assert (samples.shape[1], samples.shape[2]) == sa.sample_shape

            samples = sa.generate_samples(n_samples)

            assert samples.shape[2] == ma.input_size
            sttn = hi.states_to_numbers(np.asarray(samples.reshape(-1, ma.input_size)))
            n_s = sttn.size

            # fill in the histogram for sampler
            unique, counts = np.unique(sttn, return_counts=True)
            hist_samp = np.zeros(n_states)
            hist_samp[unique] = counts

            # expected frequencies
            f_exp = n_s * ps

            statistics, pvalues[jrep] = chisquare(hist_samp, f_exp=f_exp)

        s, pval = combine_pvalues(pvalues, method="fisher")
        assert pval > 0.01 or np.max(pvalues) > 0.01



from .abstract_optimizer import AbstractOptimizer
import numpy as _np
import torch as _torch


class Torch(AbstractOptimizer):
    r"""Wrapper for Torch optimizers."""

    def __init__(self, machine, optimizer, **opt_pars):
        r"""
        Constructs a new ``Torch`` optimizer that can be used in NetKet drivers.

        Args:
            machine (AbstractMachine): The machine to be optimized.
            optimizer (torch.optim.Optimizer): A PyTorch optimizer.
            opt_pars: named parameters to be passed to opt at construction

        Examples:
            Simple SGD optimizer from PyTorch.

            >>> from netket.optimizer import Torch
            >>> from torch.optim import SGD
            >>>
            >>> g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)
            >>> hi = nk.hilbert.Spin(s=0.5, graph=g)
            >>> ha = nk.operator.Ising(h=1.0, hilbert=hi)
            >>> ma = nk.machine.RbmReal(alpha=1, hilbert=hi)
            >>> learning_rate = 0.01
            >>> opt = Torch(ma, SGD, lr=learning_rate)
        """
        if not issubclass(optim, _torch.optim.Optimizer):
            raise ValueError("Not a valid Torch optimizer.")

        if machine.dtype is complex:
            raise ValueError(
                "Torch optimizers work only for machines with real parameters."
            )

        self._t_pars = _torch.from_numpy(machine.parameters.real)
        self._t_opt = optimizer([self._t_pars], **opt_pars)

        self.reset()

    def update(self, grad, pars):
        self._t_pars.grad = _torch.from_numpy(grad.real)
        self._t_opt.step()
        pars.real = self._t_pars.numpy()
        return pars

    def reset(self):
        self._t_opt.zero_grad()

    def __repr__(self):
        rep = "Wrapper for the following Torch optimizer :\n"
        rep += str(self._t_opt)
        return rep



from .abstract_sampler import AbstractSampler

from .metropolis_hastings import MetropolisHastings, MetropolisHastingsPt
from .metropolis_local import MetropolisLocal, MetropolisLocalPt
from .metropolis_exchange import MetropolisExchange, MetropolisExchangePt
from .metropolis_hamiltonian import MetropolisHamiltonian, MetropolisHamiltonianPt
from .custom_sampler import CustomSampler, CustomSamplerPt

from .exact_sampler import ExactSampler

from . import jax

from netket.utils import _hide_submodules

_hide_submodules(__name__)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

import numpy as np
from numba import jit

from netket.graph import AbstractGraph
from netket.utils import deprecated

from .homogeneous import HomogeneousHilbert
from ._deprecations import graph_to_N_depwarn


@jit(nopython=True)
def _sum_constraint(x, n_particles):
    return np.sum(x, axis=1) == n_particles


class Fock(HomogeneousHilbert):
    r"""Hilbert space obtained as tensor product of local fock basis."""

    def __init__(
        self,
        n_max: Optional[int] = None,
        N: int = 1,
        n_particles: Optional[int] = None,
        graph: Optional[AbstractGraph] = None,
    ):
        r"""
        Constructs a new ``Boson`` given a maximum occupation number, number of sites
        and total number of bosons.

        Args:
          n_max: Maximum occupation for a site (inclusive). If None, the local
            occupation number is unbounded.
          N: number of bosonic modes (default = 1)
          n_particles: Constraint for the number of particles. If None, no constraint
            is imposed.
          graph: (Deprecated, pleaese use `N`) A graph, from which the number of nodes
            is extracted.

        Examples:
           Simple boson hilbert space.

           >>> from netket.hilbert import Fock
           >>> hi = Fock(n_max=5, n_particles=11, N=3)
           >>> print(hi.size)
           3
           >>> print(hi.n_states)
           15
        """
        N = graph_to_N_depwarn(N=N, graph=graph)

        self._n_max = n_max

        if n_particles is not None:
            n_particles = int(n_particles)
            self._n_particles = n_particles
            assert n_particles > 0

            if self._n_max is None:
                self._n_max = n_particles
            else:
                if self._n_max * N < n_particles:
                    raise Exception(
                        """The required total number of bosons is not compatible
                        with the given n_max."""
                    )

            def constraints(x):
                return _sum_constraint(x, n_particles)

        else:
            constraints = None
            self._n_particles = None

        if self._n_max is not None:
            # assert self._n_max > 0
            local_states = np.arange(self._n_max + 1)
        else:
            max_ind = np.iinfo(np.intp).max
            self._n_max = max_ind
            local_states = None

        super().__init__(local_states, N, constraints)

    @property
    def n_max(self) -> Optional[int]:
        r"""The maximum number of bosons per site, or None
        if the number is unconstrained."""
        return self._n_max

    @property
    def n_particles(self) -> Optional[int]:
        r"""The total number of particles, or None
        if the number is unconstrained."""
        return self._n_particles

    def __pow__(self, n) -> "Fock":
        if self.n_particles is None:
            n_particles = None
        else:
            n_particles = n_particles * n

        return Fock(self.n_max, self.size * n, n_particles=n_particles)

    def _mul_sametype_(self, other: "Fock") -> "Fock":
        assert type(self) == type(other)
        if self.n_max == other.n_max:
            if self._n_particles is None and other._n_particles is None:
                return Fock(self.n_max, N=self.size + other.size)

        return NotImplemented

    def ptrace(self, sites: Union[int, List]) -> Optional["Fock"]:
        if isinstance(sites, int):
            sites = [sites]

        for site in sites:
            if site < 0 or site >= self.size:
                raise ValueError(
                    f"Site {site} not in this hilbert space of site {self.size}"
                )

        if self.n_particles is not None:
            raise TypeError(
                "Cannot take the partial trace with a total particles constraint."
            )

        Nsites = len(sites)

        if self.size - Nsites == 0:
            return None
        else:
            return Fock(self.n_max, N=self.size - Nsites)

    def __repr__(self):
        n_particles = (
            ", n_particles={}".format(self._n_particles)
            if self._n_particles is not None
            else ""
        )
        nmax = self._n_max if self._n_max < np.iinfo(np.intp).max else "INT_MAX"
        return "Fock(n_max={}{}, N={})".format(nmax, n_particles, self._size)

    @property
    def _attrs(self):
        return (self.size, self._n_max, self._constraint_fn)

    def _random_state_with_constraint_legacy(self, out, rgen, n_max):
        sites = list(range(self.size))

        out.fill(0.0)
        ss = self.size

        for _ in range(self.n_particles):
            s = rgen.integers(0, ss, size=())

            out[sites[s]] += 1

            if out[sites[s]] == n_max:
                sites.pop(s)
                ss -= 1

    def _random_state_legacy(self, size=None, *, out=None, rgen=None):
        if isinstance(size, int):
            size = (size,)
        shape = (*size, self._size) if size is not None else (self._size,)

        if out is None:
            out = np.empty(shape=shape)

        if rgen is None:
            rgen = np.random.default_rng()

        if self.n_particles is None:
            out[:] = rgen.integers(0, self.n_max, size=shape)
        else:
            if size is not None:
                out_r = out.reshape(-1, self._size)
                for b in range(out_r.shape[0]):
                    self._random_state_with_constraint_legacy(
                        out_r[b], rgen, self.n_max
                    )
            else:
                self._random_state_with_constraint_legacy(out, rgen, self.n_max)

        return out


@deprecated(
    """
Boson has been replaced by Fock. Please use fock, which has
the same semantics except for n_bosons which was replaced by
n_particles.

You should update your code because it will break in a future
version.
"""
)
def Boson(
    n_max: Optional[int] = None,
    N: int = 1,
    n_bosons: Optional[int] = None,
    graph: Optional[AbstractGraph] = None,
) -> Fock:
    return Fock(n_max, N, n_bosons, graph)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk
import numpy as np
import pytest
from scipy.stats import combine_pvalues, chisquare

import jax
import flax
from jax import numpy as jnp

from .. import common

pytestmark = common.skipif_mpi

nk.config.update("NETKET_EXPERIMENTAL", True)
np.random.seed(1234)

WEIGHT_SEED = 1234
SAMPLER_SEED = 15324


samplers = {}


# TESTS FOR SPIN HILBERT
# Constructing a 1d lattice
g = nk.graph.Hypercube(length=4, n_dim=1)

# Hilbert space of spins from given graph
hi = nk.hilbert.Spin(s=0.5, N=g.n_nodes)
ha = nk.operator.Ising(hilbert=hi, graph=g, h=1.0)
move_op = sum([nk.operator.spin.sigmax(hi, i) for i in range(hi.size)])

hi_spin1 = nk.hilbert.Spin(s=1, N=g.n_nodes)
hib = nk.hilbert.Fock(n_max=1, N=g.n_nodes, n_particles=1)
hib_u = nk.hilbert.Fock(n_max=3, N=g.n_nodes)

samplers["Exact: Spin"] = nk.sampler.ExactSampler(hi, n_chains=8)
samplers["Exact: Fock"] = nk.sampler.ExactSampler(hib_u, n_chains=4)

samplers["Metropolis(Local): Spin"] = nk.sampler.MetropolisLocal(hi, n_chains=16)

samplers["MetropolisNumpy(Local): Spin"] = nk.sampler.MetropolisLocalNumpy(
    hi, n_chains=16
)
# samplers["MetropolisNumpy(Local): Fock"] = nk.sampler.MetropolisLocalNumpy(
#    hib_u, n_chains=8
# )
# samplers["MetropolisNumpy(Local): Doubled-Spin"] = nk.sampler.MetropolisLocalNumpy(
#    nk.hilbert.DoubledHilbert(nk.hilbert.Spin(s=0.5, N=2)), n_chains=8
# )

samplers["MetropolisPT(Local): Spin"] = nk.sampler.MetropolisLocalPt(
    hi, n_chains=8, n_replicas=4
)
samplers["MetropolisPT(Local): Fock"] = nk.sampler.MetropolisLocalPt(
    hib_u, n_chains=8, n_replicas=4
)

samplers["Metropolis(Exchange): Fock-1particle"] = nk.sampler.MetropolisExchange(
    hib, n_chains=16, graph=g
)

samplers["Metropolis(Hamiltonian,Jax): Spin"] = nk.sampler.MetropolisHamiltonian(
    hi,
    hamiltonian=ha,
    reset_chains=True,
)

samplers["Metropolis(Hamiltonian,Numpy): Spin"] = nk.sampler.MetropolisHamiltonianNumpy(
    hi,
    hamiltonian=ha,
    reset_chains=True,
)

samplers["Metropolis(Custom: Sx): Spin"] = nk.sampler.MetropolisCustom(
    hi, move_operators=move_op
)

# samplers["MetropolisPT(Custom: Sx): Spin"] = nk.sampler.MetropolisCustomPt(hi, move_operators=move_op, n_replicas=4)

samplers["Autoregressive: Spin 1/2"] = nk.sampler.ARDirectSampler(hi, n_chains=16)
samplers["Autoregressive: Spin 1"] = nk.sampler.ARDirectSampler(hi_spin1, n_chains=16)
samplers["Autoregressive: Fock"] = nk.sampler.ARDirectSampler(hib_u, n_chains=16)


# The following fixture initialisees a model and it's weights
# for tests that require it.
@pytest.fixture
def model_and_weights(request):
    def build_model(hilb, sampler=None):
        if isinstance(sampler, nk.sampler.ARDirectSampler):
            ma = nk.models.ARNNDense(hilbert=hilb, layers=3, features=5)
        else:
            # Build RBM by default
            ma = nk.models.RBM(
                alpha=1,
                dtype=complex,
                kernel_init=nk.nn.initializers.normal(stddev=0.1),
                hidden_bias_init=nk.nn.initializers.normal(stddev=0.1),
            )
            # init network

        w = ma.init(jax.random.PRNGKey(WEIGHT_SEED), jnp.zeros((1, hi.size)))

        return ma, w

    # Do something with the data
    return build_model


# The following fixture returns one sampler at a time (and iterates through)
# all samplers.
#  it skips tests according to the --sampler cmdline argument introduced in
# conftest.py
@pytest.fixture(
    params=[pytest.param(sampl, id=name) for name, sampl in samplers.items()]
)
def sampler(request):
    cmdline_sampler = request.config.getoption("--sampler").lower()
    if cmdline_sampler == "":
        return request.param
    elif cmdline_sampler in request.node.name.lower():
        return request.param
    else:
        pytest.skip("skipped from command-line argument")


@pytest.fixture(params=[pytest.param(val, id=f", mpow={val}") for val in [1, 2]])
def set_pdf_power(request):
    def fun(sampler):
        cmdline_mpow = request.config.getoption("--mpow").lower()
        if cmdline_mpow == "all":
            # Nothing to skip
            pass
        elif cmdline_mpow == "single":
            # same sampler leads to same rng
            rng = np.random.default_rng(abs(hash((type(sampler), repr(sampler)))))
            exponent = rng.integers(1, 3)  # 1 or 2
            if exponent != request.param:
                pytest.skip(
                    "Running only 1 pdf exponent per sampler. Use --mpow=all to run all pdf exponents."
                )
        elif int(cmdline_mpow) != request.param:
            pytest.skip(f"Running only --mpow={cmdline_mpow}.")

        if isinstance(sampler, nk.sampler.ARDirectSampler) and request.param != 2:
            pytest.skip("ARDirectSampler only supports machine_pow = 2.")

        return sampler.replace(machine_pow=request.param)

    return fun


def test_states_in_hilbert(sampler, model_and_weights):
    hi = sampler.hilbert
    all_states = hi.all_states()

    ma, w = model_and_weights(hi, sampler)

    for sample in nk.sampler.samples(sampler, ma, w, chain_length=50):
        assert sample.shape == (sampler.n_chains, hi.size)
        for v in sample:
            assert v in all_states

    # if hasattr(sa, "acceptance"):
    #    assert np.min(sampler.acceptance) >= 0 and np.max(sampler.acceptance) <= 1.0


def findrng(rng):
    if hasattr(rng, "_bit_generator"):
        return rng._bit_generator.state["state"]
    else:
        return rng


# Mark tests that we know are failing on correctedness
def failing_test(sampler):
    if isinstance(sampler, nk.sampler.MetropolisSampler):
        if isinstance(sampler, nk.sampler.MetropolisPtSampler):
            return True

    return False


@pytest.fixture(
    params=[
        pytest.param(
            sampl,
            id=name,
            marks=pytest.mark.xfail(reason="MUSTFIX: this sampler is known to fail")
            if failing_test(sampl)
            else [],
        )
        for name, sampl in samplers.items()
    ]
)
def sampler_c(request):
    cmdline_sampler = request.config.getoption("--sampler").lower()
    if cmdline_sampler == "":
        return request.param
    elif cmdline_sampler in request.node.name.lower():
        return request.param
    else:
        pytest.skip("skipped from command-line argument")


# Testing that samples generated from direct sampling are compatible with those
# generated by markov chain sampling
# here we use a combination of power divergence tests
def test_correct_sampling(sampler_c, model_and_weights, set_pdf_power):
    sampler = set_pdf_power(sampler_c)

    hi = sampler.hilbert
    n_states = hi.n_states

    ma, w = model_and_weights(hi, sampler)

    n_samples = max(40 * n_states, 100)

    ps = np.absolute(nk.nn.to_array(hi, ma, w, normalize=False)) ** sampler.machine_pow
    ps /= ps.sum()

    n_rep = 6
    pvalues = np.zeros(n_rep)

    sampler_state = sampler.init_state(ma, w, seed=SAMPLER_SEED)

    for jrep in range(n_rep):
        sampler_state = sampler.reset(ma, w, state=sampler_state)

        # Burnout phase
        samples, sampler_state = sampler.sample(
            ma, w, state=sampler_state, chain_length=n_samples // 100
        )

        assert samples.shape == (
            n_samples // 100,
            sampler.n_chains,
            hi.size,
        )
        samples, sampler_state = sampler.sample(
            ma, w, state=sampler_state, chain_length=n_samples
        )

        assert samples.shape == (n_samples, sampler.n_chains, hi.size)

        sttn = hi.states_to_numbers(np.asarray(samples.reshape(-1, hi.size)))
        n_s = sttn.size

        # fill in the histogram for sampler
        unique, counts = np.unique(sttn, return_counts=True)
        hist_samp = np.zeros(n_states)
        hist_samp[unique] = counts

        # expected frequencies
        f_exp = n_s * ps
        statistics, pvalues[jrep] = chisquare(hist_samp, f_exp=f_exp)

    s, pval = combine_pvalues(pvalues, method="fisher")
    assert pval > 0.01 or np.max(pvalues) > 0.01


def test_throwing(model_and_weights):
    with pytest.raises(TypeError):
        nk.sampler.MetropolisHamiltonian(
            hi,
            hamiltonian=10,
            reset_chains=True,
        )

    with pytest.raises(ValueError):
        sampler = nk.sampler.MetropolisHamiltonian(
            nk.hilbert.DoubledHilbert(hi),
            hamiltonian=ha,
            reset_chains=True,
        )

        ma, w = model_and_weights(hi)

        # test raising of init state
        sampler.init_state(ma, w, seed=SAMPLER_SEED)

    with pytest.raises(ValueError):
        sampler = nk.sampler.MetropolisHamiltonianNumpy(
            nk.hilbert.Fock(3) ** hi.size,
            hamiltonian=ha,
            reset_chains=True,
        )

        ma, w = model_and_weights(hi)

        # test raising of init state
        sampler.init_state(ma, w, seed=SAMPLER_SEED)

    with pytest.raises(flax.errors.ScopeParamShapeError):
        sampler = nk.sampler.MetropolisHamiltonianNumpy(
            nk.hilbert.DoubledHilbert(hi),
            hamiltonian=ha,
            reset_chains=True,
        )

        ma, w = model_and_weights(hi)

        # test raising of init state
        sampler.init_state(ma, w, seed=SAMPLER_SEED)


def test_exact_sampler(sampler):
    known_exact_samplers = [nk.sampler.ExactSampler, nk.sampler.ARDirectSampler]
    if any(isinstance(sampler, x) for x in known_exact_samplers):
        assert sampler.is_exact is True
    else:
        assert sampler.is_exact is False



# Copyright 2020 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.legacy.vmc_common import tree_map
import pytest

pytestmark = pytest.mark.legacy


def test_tree_map():
    # Structure: name -> (input, expected)
    test_trees = {
        "None": (None, None),
        "leaf": (0, 1),
        "list": ([0, 1], [1, 2]),
        "tuple1": ((0,), (1,)),
        "tuple3": ((0, 1, 0), (1, 2, 1)),
        "list of list": ([[0, 1], [1], [0, 1, 0]], [[1, 2], [2], [1, 2, 1]]),
        "dict": ({"a": 0, "b": 1}, {"a": 1, "b": 2}),
        "mixed list": ([(0, 1), [1], 1, {"a": 0}], [(1, 2), [2], 2, {"a": 1}]),
    }

    for name, (inp, expected) in test_trees.items():
        info = "{}, input={}".format(name, inp)
        assert tree_map(lambda x: x + 1, inp) == expected, info



import math

from netket import legacy as _nk

from netket.operator import local_values as _local_values
from netket.legacy.stats import (
    statistics as _statistics,
    mean as _mean,
    sum_inplace as _sum_inplace,
)

from netket.legacy.vmc_common import info, tree_map
from netket.legacy.abstract_variational_driver import AbstractVariationalDriver


class Vmc(AbstractVariationalDriver):
    """
    Energy minimization using Variational Monte Carlo (VMC).
    """

    def __init__(
        self,
        hamiltonian,
        sampler,
        optimizer,
        n_samples,
        n_discard=None,
        sr=None,
    ):
        """
        Initializes the driver class.

        Args:
            hamiltonian (AbstractOperator): The Hamiltonian of the system.
            sampler: The Monte Carlo sampler.
            optimizer (AbstractOptimizer): Determines how optimization steps are performed given the
                bare energy gradient.
            n_samples (int): Number of Markov Chain Monte Carlo sweeps to be
                performed at each step of the optimization.
            n_discard (int, optional): Number of sweeps to be discarded at the
                beginning of the sampling, at each step of the optimization.
                Defaults to 10% of the number of samples allocated to each MPI node.
            sr (SR, optional): Determines whether and how stochastic reconfiguration
                is applied to the bare energy gradient before performing applying
                the optimizer. If this parameter is not passed or None, SR is not used.

        Example:
            Optimizing a 1D wavefunction with Variational Monte Carlo.

            >>> import netket as nk
            >>> SEED = 3141592
            >>> g = nk.graph.Hypercube(length=8, n_dim=1)
            >>> hi = nk.hilbert.Spin(s=0.5, graph=g)
            >>> ma = nk.machine.RbmSpin(hilbert=hi, alpha=1)
            >>> ma.init_random_parameters(seed=SEED, sigma=0.01)
            >>> ha = nk.operator.Ising(hi, h=1.0)
            >>> sa = nk.sampler.MetropolisLocal(machine=ma)
            >>> op = nk.optimizer.Sgd(learning_rate=0.1)
            >>> vmc = nk.Vmc(ha, sa, op, 200)

        """
        super(Vmc, self).__init__(
            sampler.machine, optimizer, minimized_quantity_name="Energy"
        )

        self._ham = hamiltonian
        self._sampler = sampler
        self.sr = sr

        self._npar = self._machine.n_par

        self._batch_size = sampler.sample_shape[0]

        # Check how many parallel nodes we are running on
        self.n_nodes = _nk.utils.mpi.n_nodes

        self.n_samples = n_samples
        self.n_discard = n_discard

        self._dp = None

    @property
    def sr(self):
        return self._sr

    @sr.setter
    def sr(self, sr):
        self._sr = sr
        if self._sr is not None:
            self._sr.setup(self.machine)

    @property
    def n_samples(self):
        return self._n_samples

    @n_samples.setter
    def n_samples(self, n_samples):
        if n_samples <= 0:
            raise ValueError(
                "Invalid number of samples: n_samples={}".format(n_samples)
            )

        n_samples_chain = int(math.ceil((n_samples / self._batch_size)))
        self._n_samples_node = int(math.ceil(n_samples_chain / self.n_nodes))

        self._n_samples = int(self._n_samples_node * self._batch_size * self.n_nodes)

        self._samples = None

        self._grads = None
        self._jac = None

    @property
    def n_discard(self):
        return self._n_discard

    @n_discard.setter
    def n_discard(self, n_discard):
        if n_discard is not None and n_discard < 0:
            raise ValueError(
                "Invalid number of discarded samples: n_discard={}".format(n_discard)
            )
        self._n_discard = (
            int(n_discard)
            if n_discard != None
            else self._n_samples_node * self._batch_size // 10
        )

    def _forward_and_backward(self, sample=True):
        """
        Performs a number of VMC optimization steps.

        Args:
            n_steps (int): Number of steps to perform.
        """

        if sample:
            self._sampler.reset()

            # Burnout phase
            self._sampler.generate_samples(self._n_discard)

            # Generate samples and store them
            self._samples = self._sampler.generate_samples(
                self._n_samples_node, samples=self._samples
            )

        # Compute the local energy estimator and average Energy
        eloc, self._loss_stats = self._get_mc_stats(self._ham)

        # Center the local energy
        eloc -= _mean(eloc)

        samples_r = self._samples.reshape((-1, self._samples.shape[-1]))
        eloc_r = eloc.reshape(-1, 1)

        # Perform update
        if self._sr:
            if self._sr.onthefly:

                self._grads = self._machine.vector_jacobian_prod(
                    samples_r, eloc_r / self._n_samples, self._grads
                )

                self._grads = tree_map(_sum_inplace, self._grads)

                self._dp = self._sr.compute_update_onthefly(
                    samples_r, self._grads, self._dp
                )

            else:
                # When using the SR (Natural gradient) we need to have the full jacobian
                self._grads, self._jac = self._machine.vector_jacobian_prod(
                    samples_r,
                    eloc_r / self._n_samples,
                    self._grads,
                    return_jacobian=True,
                )

                self._grads = tree_map(_sum_inplace, self._grads)

                self._dp = self._sr.compute_update(self._jac, self._grads, self._dp)

        else:
            # Computing updates using the simple gradient
            self._grads = self._machine.vector_jacobian_prod(
                samples_r, eloc_r / self._n_samples, self._grads
            )

            self._grads = tree_map(_sum_inplace, self._grads)

            #  if Real pars but complex gradient, take only real part
            # not necessary for SR because sr already does it.
            if not self._machine.has_complex_parameters:
                self._dp = tree_map(lambda x: x.real, self._grads)
            else:
                self._dp = self._grads

        return self._dp

    @property
    def energy(self):
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.
        """
        return self._loss_stats

    def _estimate_stats(self, obs):
        if self._samples is None:
            raise RuntimeError(
                "Vmc driver needs to perform a step before .estimate() can be "
                "called. To get VMC estimates outside of optimization, use "
                "netket.vqs.estimate_expectations instead."
            )
        return self._get_mc_stats(obs)[1]

    def reset(self):
        self._sampler.reset()
        super().reset()

    def _get_mc_stats(self, op):

        samples_r = self._samples.reshape((-1, self._samples.shape[-1]))

        loc = _local_values(op, self._machine, samples_r).reshape(
            self._samples.shape[0:2]
        )

        # notice that loc.T is passed to statistics, since that function assumes
        # that the first index is the batch index.
        return loc, _statistics(loc.T)

    def __repr__(self):
        return "Vmc(step_count={}, n_samples={}, n_discard={})".format(
            self.step_count, self.n_samples, self.n_discard
        )

    def info(self, depth=0):
        lines = [
            "{}: {}".format(name, info(obj, depth=depth + 1))
            for name, obj in [
                ("Hamiltonian ", self._ham),
                ("Machine     ", self._machine),
                ("Optimizer   ", self._optimizer),
                ("SR solver   ", self._sr),
            ]
        ]
        return "\n{}".format(" " * 3 * (depth + 1)).join([str(self)] + lines)



from functools import singledispatch
from . import numpy


@singledispatch
def MetropolisHastings(machine, kernel, n_chains=16, sweep_size=None):
    r"""
    ``MetropolisHastings`` is a generic Metropolis-Hastings sampler using
    a transition kernel to perform moves in the Markov Chain.
    The transition kernel is used to generate
    a proposed state :math:`s^\prime`, starting from the current state :math:`s`.
    The move is accepted with probability

    .. math::
    A(s\rightarrow s^\prime) = \mathrm{min}\left (1,\frac{P(s^\prime)}{P(s)} F(e^{L(s,s^\prime)})\right),

    where the probability being sampled from is :math:`P(s)=|M(s)|^p. Here ::math::`M(s)` is a
    user-provided function (the machine), :math:`p` is also user-provided with default value :math:`p=2`,
    and :math:`L(s,s^\prime)` is a suitable correcting factor computed by the transition kernel.


    Args:
        machine: A machine :math:`M(s)` used for the sampling.
                The probability distribution being sampled
                from is :math:`P(s)=|M(s)|^p`, where the power :math:`p`
                is arbitrary, and by default :math:`p=2`.
        kernel: A kernel to generate random transitions from a given state as
                well as uniform random states.
                This must have two methods, `random_state` and `transition`.
                `random_state` takes an input state (in batches) and
                changes it in-place to a valid random state.
                `transition` takes an input state (in batches) and
                returns a batch of random states obtained transitioning from the initial state.
                `transition` must also return an array containing the
                `log_prob_corrections` :math:`L(s,s^\prime)`.
        n_chains: The number of Markov Chain to be run in parallel on a single process.
        sweep_size: The number of exchanges that compose a single sweep.
                If None, sweep_size is equal to the number of degrees of freedom being sampled
                (the size of the input vector s to the machine).

    """

    return numpy.MetropolisHastings(machine, kernel, n_chains, sweep_size)


@singledispatch
def MetropolisHastingsPt(machine, kernel, n_replicas=32, sweep_size=None):
    r"""
    ``MetropolisHastingsPt`` is a generic Metropolis-Hastings sampler using
    a local transition kernel to perform moves in the Markov Chain and replica-exchange moves
    to increase mixing times.
    The transition kernel is used to generate a proposed state :math:`s^\prime`,
    starting from the current state :math:`s`, as in ``MetropolisHastings``.
    Each replica is at an inverse temperature :math:`\beta` linearly spaced in the interval [0,1],
    where :math:`beta=1` corresponds to the target probability distribution,
    and :math:`beta=0` is the flat distribution.

    Args:
        machine: A machine :math:`M(s)` used for the sampling.
                The probability distribution being sampled
                from is :math:`P(s)=|M(s)|^p`, where the power :math:`p`
                is arbitrary, and by default :math:`p=2`.
        kernel: A kernel to generate random transitions from a given state as
                well as uniform random states.
                This must have two methods, `random_state` and `transition`.
                `random_state` takes an input state (in batches) and
                changes it in-place to a valid random state.
                `transition` takes an input state (in batches) and
                returns a batch of random states obtained transitioning from the initial state.
                `transition` must also return an array containing the
                `log_prob_corrections` :math:`L(s,s^\prime)`.
        n_replicas (int): The number of replicas used for replica-exchange moves. Each replica samples
        sweep_size (int): The number of exchanges that compose a single sweep.
                    If None, sweep_size is equal to the number of degrees of freedom (the input size of the machine).

    """
    return numpy.MetropolisHastingsPt(machine, kernel, n_replicas, sweep_size)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import jit

from netket.legacy.machine import Jax as _Jax

from ._der_local_values_jax import der_local_values_jax


@jit(nopython=True)
def _der_local_values_kernel(
    log_vals, log_val_p, mels, der_log, der_log_p, sections, out
):
    low_range = 0
    for i, s in enumerate(sections):
        out[i, :] = (
            np.expand_dims(
                mels[low_range:s] * np.exp(log_val_p[low_range:s] - log_vals[i]), 1
            )
            * (der_log_p[low_range:s, :] - der_log[i, :])
        ).sum(axis=0)
        low_range = s


def _der_local_values_impl(op, machine, v, log_vals, der_log_vals, out, batch_size=64):
    sections = np.empty(v.shape[0], dtype=np.int32)
    v_primes, mels = op.get_conn_flattened(v, sections)

    log_val_primes = machine.log_val(v_primes)

    # Compute the der_log in small batches and not in one go.
    # For C++ machines there is a 100% slowdown when the batch is too big.
    n_primes = len(log_val_primes)
    der_log_primes = np.empty((n_primes, machine.n_par), dtype=np.complex128)

    for s in range(0, n_primes, batch_size):
        end = min(s + batch_size, n_primes)
        der_log_primes[s:end, :] = machine.der_log(v_primes[s:end])

    _der_local_values_kernel(
        log_vals, log_val_primes, mels, der_log_vals, der_log_primes, sections, out
    )


@jit(nopython=True)
def _der_local_values_notcentered_kernel(
    log_vals, log_val_p, mels, der_log_p, sections, out
):
    low_range = 0
    for i, s in enumerate(sections):
        out[i, :] = (
            np.expand_dims(
                mels[low_range:s] * np.exp(log_val_p[low_range:s] - log_vals[i]), 1
            )
            * der_log_p[low_range:s, :]
        ).sum(axis=0)
        low_range = s


def _der_local_values_notcentered_impl(op, machine, v, log_vals, out, batch_size=64):
    sections = np.empty(v.shape[0], dtype=np.int32)
    v_primes, mels = op.get_conn_flattened(v, sections)

    log_val_primes = machine.log_val(v_primes)

    # Compute the der_log in small batches and not in one go.
    # For C++ machines there is a 100% slowdown when the batch is too big.
    n_primes = len(log_val_primes)
    der_log_primes = np.empty((n_primes, machine.n_par), dtype=np.complex128)

    for s in range(0, n_primes, batch_size):
        end = min(s + batch_size, n_primes)
        der_log_primes[s:end, :] = machine.der_log(v_primes[s:end])

    _der_local_values_notcentered_kernel(
        log_vals, log_val_primes, mels, der_log_primes, sections, out
    )


def der_local_values(
    op,
    machine,
    v,
    log_vals=None,
    der_log_vals=None,
    out=None,
    center_derivative=True,
    batch_size=64,
):
    r"""
    Computes the derivative of local values of the operator `op` for all `samples`.

    The local value is defined as
    .. math:: O_{\mathrm{loc}}(x) = \langle x | O | \Psi \rangle / \langle x | \Psi \rangle


            Args:
                op: Hermitian operator.
                v: A numpy array or matrix containing either a single
                    :math:`V = v` or a batch of visible
                    configurations :math:`V = v_1,\dots v_M`.
                    In the latter case, each row of the matrix corresponds to a
                    visible configuration.
                machine: Wavefunction :math:`\Psi`.
                log_vals: A scalar/numpy array containing the value(s) :math:`\Psi(V)`.
                    If not given, it is computed from scratch.
                    Defaults to None.
                der_log_vals: A numpy tensor containing the vector of log-derivative(s) :math:`O_i(V)`.
                    If not given, it is computed from scratch.
                    Defaults to None.
                out: A scalar or a numpy array of local values of the operator.
                    If not given, it is allocated from scratch and then returned.
                    Defaults to None.
                center_derivative: Whever to center the derivatives or not. In the formula above,
                    When this is true/false it is equivalent to setting :math:`\alpha=\{1 / 2\}`.
                    By default `center_derivative=True`, meaning that it returns the correct
                    derivative of the local values. False is mainly used when dealing with liouvillians.

            Returns:
                If samples is given in batches, a numpy ndarray of derivatives of local values
                of the operator, otherwise a 1D array.
    """
    if isinstance(machine, _Jax):
        return der_local_values_jax(
            op,
            machine,
            v,
            log_vals=log_vals,
            center_derivative=center_derivative,
        )

    if v.ndim != 2:
        raise RuntimeError("Invalid input shape, expected a 2d array")

    assert (
        v.shape[1] == op.hilbert.size
    ), "samples has wrong shape: {}; expected (?, {})".format(v.shape, op.hilbert.size)

    if out is None:
        out = np.empty((v.shape[0], machine.n_par), dtype=np.complex128)

    if log_vals is None:
        log_vals = machine.log_val(v)

    if der_log_vals is None and center_derivative is True:
        der_log_vals = machine.der_log(v)

    if center_derivative is True:
        _der_local_values_impl(
            op, machine, v, log_vals, der_log_vals, out, batch_size=batch_size
        )
    else:
        _der_local_values_notcentered_impl(
            op, machine, v, log_vals, out, batch_size=batch_size
        )

    return out



# Copyright 2020 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_machine import AbstractMachine
from netket.graph import AbstractGraph
import numpy as _np
from numba import jit


class Jastrow(AbstractMachine):
    r"""
    A Jastrow wavefunction Machine. This machine defines the following
    wavefunction:

    .. math:: \Psi(s_1,\dots s_N) = e^{\sum_{ij} s_i J_{ij} s_j + \sum_{i} a_i s_i}

    where :math:` W_{ij}` are the two-body Jastrow parameters and a_i are optional
    visible bias (single-body effective potential).

    Symmetries in the weights J_{i,j} can be imposed passing a valid automorphism.

    The weights can be taken to be complex-valued (default option) or real-valued.
    """

    def __init__(
        self,
        hilbert,
        use_visible_bias=False,
        automorphisms=None,
        dtype=complex,
    ):
        r"""
        Constructs a new ``Jastrow`` machine:

        Args:
            hilbert: Hilbert space object for the system.
            use_visible_bias (bool): Whether to use the visible bias a_i.
            automorphisms (optional): Can be a graph or a custom matrix of automorphisms.
            dtype: either complex or float, it is the type used for the weights and biases.

        Examples:
            A ``Jastrow`` machine for a one-dimensional L=20 spin 1/2
            system:

            >>> from netket.machine import Jastrow
            >>> from netket.hilbert import Spin
            >>> hi = Spin(s=0.5, total_sz=0, N=20)
            >>> ma = Jastrow(hilbert=hi)
            >>> print(ma.n_par)
            190
        """

        n = hilbert.size

        if dtype is not float and dtype is not complex:
            raise TypeError("dtype must be either float or complex")

        self._dtype = dtype
        self._npdtype = _np.complex128 if dtype is complex else _np.float64

        if n < 2:
            raise ValueError(
                "Cannot construct a Jastrow factor with less than two visible units."
            )

        if automorphisms is None:
            n_sym = int((n * (n - 1)) // 2)
            self._Smap = _np.zeros((n, n), dtype=_np.intp)

            k = 0
            for i in range(n):
                for j in range(i + 1, n, 1):
                    self._Smap[i, j] = k
                    self._Smap[j, i] = k
                    k += 1

        else:
            if isinstance(automorphisms, AbstractGraph):
                autom = _np.asarray(automorphisms.automorphisms())
            else:
                try:
                    autom = _np.asarray(automorphisms)
                    assert n == autom.shape[1]
                except:
                    raise RuntimeError("Cannot find a valid automorphism array.")

            self._Smap, n_sym = self._gen_symm(autom)

        # The symmetric part of J is stored in a 1d array
        self._J = _np.zeros(n_sym, dtype=self._npdtype)

        # Visible bias
        self._a = _np.empty(n, dtype=self._npdtype) if use_visible_bias else None

        self._npar = n_sym + (self._a.size if self._a is not None else 0)

        super().__init__(hilbert, dtype)

    @property
    def n_par(self):
        r"""The number of variational parameters in the machine."""
        return self._npar

    def log_val(self, x, out=None):
        r"""Computes the logarithm of the wave function for a batch of visible
        configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. The
                 length of `out` should be `x.shape[0]`.

        Returns:
            A complex number when `x` is a vector and vector when `x` is a
            matrix.
        """
        x = x.astype(dtype=self._npdtype)

        return self._log_val_kernel(x, out, self._J, self._a, self._Smap)

    @staticmethod
    @jit(nopython=True)
    def _log_val_kernel(x, out, J, a, Smap):
        if out is None:
            out = _np.empty(x.shape[0], dtype=_np.complex128)

        for b in range(x.shape[0]):
            out[b] = 0
            for i in range(x.shape[1]):
                for j in range(i + 1, x.shape[1], 1):
                    out[b] += x[b, i] * J[Smap[i, j]] * x[b, j]

        if a is not None:
            out = out + x.dot(a)

        return out

    def der_log(self, x, out=None):
        r"""Computes the gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and stores the result into `out`.

        Args:
            x: A matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination tensor of `complex128`.
                `out` should be a matrix of shape `(v.shape[0], self.n_par)`.

        Returns:
            `out`
        """
        x = x.astype(dtype=self._npdtype)

        return self._der_log_kernel(x, out, self._a, self._J, self._npar, self._Smap)

    @staticmethod
    @jit(nopython=True)
    def _der_log_kernel(x, out, a, J, n_par, Smap):
        batch_size = x.shape[0]

        if out is None:
            out = _np.empty((batch_size, n_par), dtype=_np.complex128)

        out.fill(0.0)

        n = x.shape[1]

        if a is not None:
            out[:, 0:n] = x
            k = n
        else:
            k = 0

        for b in range(batch_size):
            for i in range(n):
                for j in range(i + 1, n, 1):
                    out[b, Smap[i, j] + k] += x[b, i] * x[b, j]

        return out

    @property
    def state_dict(self):
        r"""A dictionary containing the parameters of this machine"""
        from collections import OrderedDict

        od = OrderedDict()
        if self._dtype is complex:
            if self._a is not None:
                od["a"] = self._a.view()

            od["J"] = self._J.view()

        else:
            if self._a is not None:
                self._ac = self._a.astype(_np.complex128)
                self._a = self._ac.real.view()
                od["a"] = self._ac.view()

            self._Jc = self._J.astype(_np.complex128)
            self._J = self._Jc.real.view()
            od["J"] = self._Jc.view()

        return od

    @staticmethod
    @jit(nopython=True)
    def _gen_symm(am):
        n = am.shape[1]

        Wt = _np.zeros((n, n), dtype=_np.intp)

        k = 0
        for i in range(n):
            for j in range(i + 1, n, 1):
                for l in range(am.shape[0]):
                    isymm = am[l, i]
                    jsymm = am[l, j]

                    Wt[isymm, jsymm] = k
                    Wt[jsymm, isymm] = k

                k += 1

        nk_unique = 0
        pars_d = dict()
        for i in range(n):
            for j in range(i + 1, n, 1):
                k = Wt[i, j]
                if k not in pars_d:
                    pars_d[k] = nk_unique
                    nk_unique += 1

        npar = len(pars_d)

        for i in range(n):
            for j in range(i + 1, n, 1):
                k = Wt[i, j]
                if k in pars_d:
                    Wt[i, j] = pars_d[k]
                    Wt[j, i] = Wt[i, j]
                else:
                    raise RuntimeError("Error constructing the symmetry table.")

        return Wt, npar


class JastrowSymm(Jastrow):
    r"""
    A Jastrow wavefunction Machine with lattice symmetries.This machine
    defines the wavefunction as follows:

    .. math:: \Psi(s_1,\dots s_N) = e^{\sum_{ij} s_i W_{ij} s_j}

    where :math:` W_{ij}` are the Jastrow parameters respects the
    specified symmetries of the lattice.

    This is just a convenience synonim for netket.machine.Jastrow with argument
    symmetry=True.
    """

    def __init__(self, hilbert, automorphisms, use_visible_bias=False, dtype=complex):
        return super().__init__(
            hilbert=hilbert,
            use_visible_bias=use_visible_bias,
            automorphisms=automorphisms,
            dtype=dtype,
        )
